"""test gini2gis"""

from unittest.mock import patch

from pyiem.nws import gini
from pyiem.util import utc

import pywwa
from pywwa.testing import get_example_filepath
from pywwa.workflows import gini2gis


def make_gini():
    """Generate a GINI file."""
    with open(get_example_filepath("TIGH05"), "rb") as fh:
        return gini.GINIZFile(fh)


def test_process_input():
    """Test process_input by mocking sys.stdin"""
    with patch("sys.stdin.buffer.read") as mock_read:
        with open(get_example_filepath("TIGH05"), "rb") as fh:
            mock_read.return_value = fh.read()
        assert gini2gis.process_input() is not None


def test_workflow():
    """Test the workflow."""
    pywwa.CTX["utcnow"] = utc(2021, 9, 16, 18)
    gini2gis.process_input = make_gini
    gini2gis.workflow()



"""Convert noaaport GINI imagery into GIS PNGs

I convert raw GINI noaaport imagery into geo-referenced PNG files both in the
'native' projection and 4326.
"""

import json
import os
import subprocess
import sys
import tempfile
from io import BytesIO

import click
import numpy as np
from PIL import Image
from pyiem.nws import gini

from pywwa import LOG, common

WORLDFILE_FORMAT = "%(dx).3f\n0.0\n0.0\n-%(dy).3f\n%(x0).3f\n%(y1).3f"


def process_input():
    """
    Process what was provided to use by LDM on stdin
    @return GINIZFile instance
    """
    cstr = BytesIO()
    payload = getattr(sys.stdin, "buffer", sys.stdin).read()
    if len(payload) < 100:
        return None
    cstr.write(payload)
    cstr.seek(0)
    return gini.GINIZFile(cstr)


def do_legacy_ir(sat, tmpfn):
    """since some are unable to process non-grayscale"""
    png = Image.fromarray(np.array(sat.data[:-1, :], np.uint8))
    png.save(f"{tmpfn}.png")

    # World File

    with open(f"{tmpfn}.wld", "w") as fh:
        fh.write(WORLDFILE_FORMAT % sat.metadata)

    tstamp = sat.metadata["valid"].strftime("%Y%m%d%H%M")
    cmd = [
        "pqinsert",
        "-i",
        "-p",
        f"gis c {tstamp} gis/images/awips{sat.awips_grid()}/"
        f"{sat.current_filename().replace('.png', '_gray.png')} GIS/sat/"
        f"awips{sat.awips.grid()}/{sat.archive_filename()} png",
        f"{tmpfn}.png",
    ]
    subprocess.call(cmd)

    cmd = [
        "pqinsert",
        "-i",
        "-p",
        f"gis c {tstamp} gis/images/awips{sat.awips_grid()}/"
        f"{sat.current_filename().replace('.png', '_gray.wld')} GIS/sat/"
        f"awips{sat.awips.grid()}/{sat.archive_filename()} wld",
        f"{tmpfn}.wld",
    ]
    subprocess.call(cmd)


def write_gispng(sat, tmpfn):
    """
    Write PNG file with associated .wld file.
    """

    if sat.get_channel() == "IR":
        do_legacy_ir(sat, tmpfn)
        # Make enhanced version
        png = Image.fromarray(np.array(sat.data[:-1, :], np.uint8))
        png.putpalette(tuple(gini.get_ir_ramp().ravel()))
    else:
        png = Image.fromarray(np.array(sat.data[:-1, :], np.uint8))
    png.save(f"{tmpfn}.png")

    # World File
    with open(f"{tmpfn}.wld", "w") as fh:
        fh.write(WORLDFILE_FORMAT % sat.metadata)

    tstamp = sat.metadata["valid"].strftime("%Y%m%d%H%M")
    cmd = [
        "pqinsert",
        "-i",
        "-p",
        f"gis {get_ldm_routes(sat)} {tstamp} gis/images/"
        f"awips{sat.awips_grid()}/{sat.current_filename()} "
        f"GIS/sat/awips{sat.awips_grid()}/{sat.archive_filename()} png",
        f"{tmpfn}.png",
    ]
    subprocess.call(cmd)

    cmd[3] = cmd[3].replace("png", "wld")
    cmd[4] = cmd[4].replace("png", "wld")
    subprocess.call(cmd)


def write_metadata(sat, tmpfn):
    """
    Write a JSON formatted metadata file
    """
    metadata = {"meta": {}}
    tstamp = sat.metadata["valid"].strftime("%Y-%m-%dT%H:%M:%SZ")
    metadata["meta"]["valid"] = tstamp
    metadata["meta"]["awips_grid"] = sat.awips_grid()
    metadata["meta"]["bird"] = sat.get_bird()
    metadata["meta"]["archive_filename"] = sat.archive_filename()
    metafp = f"{tmpfn}.json"
    with open(metafp, "w") as fh:
        json.dump(metadata, fh)

    tstamp = sat.metadata["valid"].strftime("%Y%m%d%H%M")
    cmd = [
        "pqinsert",
        "-i",
        "-p",
        f"gis c {tstamp} gis/images/awips{sat.awips_grid()}/"
        f"{sat.current_filename().replace('png', 'json')} "
        f"GIS/sat/{sat.archive_filename().replace('png', 'json')} json",
        f"{tmpfn}.json",
    ]
    subprocess.call(cmd)
    os.unlink(f"{tmpfn}.json")


def write_mapserver_metadata(sat, tmpfn, epsg):
    """Write out and pqinsert a metadata file that mapserver can use to
    provide WMS metadata."""
    metafn = f"{tmpfn}.txt"
    with open(metafn, "w") as fh:
        fh.write(
            f"""
  METADATA
    "wms_title" "{sat.get_bird()} {sat.get_sector()} {sat.get_channel()} \
 valid {sat.metadata["valid"]:%Y-%m-%dT%H:%M:%SZ} UTC"
    "wms_srs"   "EPSG:4326 EPSG:26915 EPSG:900913 EPSG:3857"
    "wms_extent" "-126 24 -66 50"
  END
"""
        )
    cmd = [
        "pqinsert",
        "-i",
        "-p",
        f"gis c {sat.metadata['valid']:%Y%m%d%H%M} "
        f"gis/images/{epsg}/goes/"
        f"{sat.current_filename().replace('png', 'msinc')} bogus msinc",
        metafn,
    ]
    subprocess.call(cmd)
    os.unlink(metafn)


def write_metadata_epsg(sat, tmpfn, epsg):
    """
    Write a JSON formatted metadata file
    """
    metadata = {"meta": {}}
    tstamp = sat.metadata["valid"].strftime("%Y-%m-%dT%H:%M:%SZ")
    metadata["meta"]["valid"] = tstamp
    metadata["meta"]["epsg"] = epsg
    metadata["meta"]["bird"] = sat.get_bird()
    metadata["meta"]["archive_filename"] = sat.archive_filename()
    metafp = f"{tmpfn}_{epsg}.json"
    with open(metafp, "w") as fh:
        json.dump(metadata, fh)

    tstamp = sat.metadata["valid"].strftime("%Y%m%d%H%M")
    cmd = [
        "pqinsert",
        "-i",
        "-p",
        f"gis c {tstamp} gis/images/{epsg}/goes/"
        f"{sat.current_filename().replace('png', 'json')} bogus json",
        f"{tmpfn}_{epsg}.json",
    ]
    subprocess.call(cmd)
    os.unlink(f"{tmpfn}_{epsg}.json")


def get_ldm_routes(sat):
    """
    Figure out if this product should be routed to current or archived folders
    """
    minutes = (common.utcnow() - sat.metadata["valid"]).seconds / 60.0
    if minutes > 120:
        return "a"
    return "ac"


def gdalwarp(sat, tmpfn, epsg):
    """
    Convert imagery into some EPSG projection, typically 4326
    """
    cmd = [
        "gdalwarp",
        "-q",
        "-of",
        "GTiff",
        "-co",
        "TFW=YES",
        "-s_srs",
        sat.metadata["proj"].srs,
        "-t_srs",
        f"EPSG:{epsg}",
        f"{tmpfn}.png",
        f"{tmpfn}_{epsg}.tif",
    ]
    subprocess.call(cmd)

    # Convert file back to PNG for use and archival (smaller file)
    cmd = [
        "convert",
        "-quiet",
        "-define",
        "PNG:preserve-colormap",
        f"{tmpfn}_{epsg}.tif",
        f"{tmpfn}_4326.png",
    ]
    proc = subprocess.Popen(
        cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    )
    output = proc.stderr.read()
    if output != b"":
        LOG.error("gdalwarp() convert error message: %s", output)
    os.unlink(f"{tmpfn}_{epsg}.tif")

    tstamp = sat.metadata["valid"].strftime("%Y%m%d%H%M")
    cmd = [
        "pqinsert",
        "-i",
        "-p",
        f"gis c {tstamp} gis/images/{epsg}/goes/"
        f"{sat.current_filename().replace('png', 'wld')} bogus wld",
        f"{tmpfn}_{epsg}.tfw",
    ]
    subprocess.call(cmd)

    cmd[3] = cmd[3].replace("wld", "png")
    cmd[4] = cmd[4].replace("wld", "png")
    subprocess.call(cmd)

    os.unlink(f"{tmpfn}_{epsg}.png")
    os.unlink(f"{tmpfn}_{epsg}.tfw")


def cleanup(tmpfn):
    """
    Pickup after ourself
    """
    for suffix in ["png", "wld", "tfw"]:
        if os.path.isfile(f"{tmpfn}.{suffix}"):
            os.unlink(f"{tmpfn}.{suffix}")


def workflow():
    """Go."""
    sat = process_input()
    if sat is None:
        LOG.info("ABORT: No data found!")
        return
    LOG.info("Processed archive file: %s", sat.archive_filename())
    if sat.awips_grid() is None:
        LOG.info("ABORT: Unknown awips grid!")
        return

    # Generate a temporary filename to use for our work
    with tempfile.NamedTemporaryFile() as tmpfd:
        # Write PNG
        write_gispng(sat, tmpfd.name)

        if get_ldm_routes(sat) == "ac":
            # Write JSON metadata
            write_metadata(sat, tmpfd.name)
            # Write JSON metadata for 4326 file
            write_metadata_epsg(sat, tmpfd.name, 4326)
            # write mapserver include metadatab
            write_mapserver_metadata(sat, tmpfd.name, 4326)
            # Warp the file into 4326
            gdalwarp(sat, tmpfd.name, 4326)
        # cleanup after ourself
        cleanup(tmpfd.name)

    LOG.info("Done!")


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """workflow"""
    workflow()



"""LSR product ingestor"""

# Stdlib
import datetime
import os
import pickle

# 3rd Party
import click
from pyiem import reference
from pyiem.nws.products.lsr import parser as lsrparser
from pyiem.util import utc
from twisted.internet import reactor

# Local
from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge

# Cheap datastore for LSRs to avoid Dups!
LSRDB = {}


def loaddb():
    """load memory"""
    if os.path.isfile("lsrdb.p"):
        with open("lsrdb.p", "rb") as fh:
            mydict = pickle.load(fh)
        for key in mydict:
            LSRDB[key] = mydict[key]


def cleandb():
    """To keep LSRDB from growing too big, we clean it out
    Lets hold 7 days of data!
    """
    now = utc()
    thres = now - datetime.timedelta(hours=24 * 7)
    init_size = len(LSRDB)
    # loop safety here
    for key in list(LSRDB):
        if LSRDB[key] < thres:
            LSRDB.pop(key)

    fin_size = len(LSRDB)
    LOG.info("cleandb() init_size: %s final_size: %s", init_size, fin_size)
    # Non blocking hackery
    reactor.callInThread(pickledb)

    # Call Again in 30 minutes
    reactor.callLater(60 * 30, cleandb)


def pickledb():
    """Dump our database to a flat file"""
    with open("lsrdb.p", "wb") as fh:
        pickle.dump(LSRDB, fh)


def real_processor(txn, text):
    """Lets actually process!"""
    prod = lsrparser(text, ugc_provider={})

    for lsr in prod.lsrs:
        if lsr.typetext.upper() not in reference.lsr_events:
            errmsg = f"Unknown LSR typecode '{lsr.typetext}'"
            common.email_error(errmsg, text)
        uniquekey = hash(lsr.text)
        if uniquekey in LSRDB:
            prod.duplicates += 1
            # akrherz/pyWWA#150, This will modify database write below
            lsr.duplicate = True
        LSRDB[uniquekey] = utc()
        if common.dbwrite_enabled():
            lsr.sql(txn)

    j = prod.get_jabbers(common.SETTINGS.get("pywwa_lsr_url", "pywwa_lsr_url"))
    for i, (p, h, x) in enumerate(j):
        # delay some to perhaps stop triggering SPAM lock outs at twitter
        reactor.callLater(i, common.send_message, p, h, x)

    if prod.warnings:
        common.email_error("\n\n".join(prod.warnings), text)
    elif not prod.lsrs:
        raise Exception("No LSRs parsed!", text)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    reactor.callLater(0, loaddb)
    bridge(real_processor, dbpool=get_database("postgis"))
    reactor.callLater(20, cleandb)



"""Database utilities."""

# 3rd Party
from psycopg.rows import dict_row
from pyiem.database import get_dbconn as pyiem_get_dbconn
from pyiem.database import get_dbconnc as pyiem_get_dbconnc
from pyiem.nws import nwsli
from twisted.enterprise import adbapi

# Local
from pywwa import LOG, SETTINGS


def get_dbconnc(dbname):
    """wrapper to pyiem to get the database connection right."""
    opts = SETTINGS.get("dbxref", {}).get(dbname, {})
    return pyiem_get_dbconnc(
        database=opts.get("database", dbname),
        host=opts.get("host", f"iemdb-{dbname}.local"),
        user=opts.get("user", "ldm"),
        port=opts.get("port", 5432),
    )


def get_dbconn(dbname):
    """wrapper to pyiem to get the database connection right."""
    opts = SETTINGS.get("dbxref", {}).get(dbname, {})
    return pyiem_get_dbconn(
        database=opts.get("database", dbname),
        host=opts.get("host", f"iemdb-{dbname}.local"),
        user=opts.get("user", "ldm"),
        port=opts.get("port", 5432),
    )


def get_database(dbname, cp_max=1, module_name="psycopg"):
    """Get a twisted database connection

    Args:
      dbname (str): The string name of the database to connect to
      cp_max (int): The maximum number of connections to make to the database
      module_name (str): The python module to use for the ConnectionPool
    """
    opts = SETTINGS.get("dbxref", {}).get(dbname, {})
    return adbapi.ConnectionPool(
        module_name,
        dbname=opts.get("database", dbname),
        cp_reconnect=True,
        cp_max=cp_max,
        host=opts.get("host", f"iemdb-{dbname}.local"),
        user=opts.get("user", "ldm"),
        port=opts.get("port", 5432),
        gssencmode="disable",  # NOTE: this is problematic with older postgres
        row_factory=dict_row,
    )


def load_nwsli(nwsli_dict):
    """Synchronous load of metadata tables."""
    with get_dbconn("postgis") as pgconn:
        cursor = pgconn.cursor(row_factory=dict_row)
        cursor.execute(
            "SELECT nwsli, river_name || ' ' || proximity || ' ' || "
            "name || ' ['||state||']' as rname from hvtec_nwsli"
        )
        for row in cursor:
            nm = row["rname"].replace("&", " and ")
            nwsli_dict[row["nwsli"]] = nwsli.NWSLI(row["nwsli"], name=nm)

        LOG.info("nwsli_dict loaded %s entries", len(nwsli_dict))


def load_metar_stations(txn, nwsli_provider):
    """load station metadata to build a xref of stations to networks"""
    # if tzname is null, this is likely some newly added station that needs
    # more metadata established prior to full usage
    txn.execute(
        """
        SELECT id, iemid, network, tzname, wfo, state, name,
        ST_X(geom) as lon, ST_Y(geom) as lat from stations
        where network ~* 'ASOS' and tzname is not null
        """
    )
    news = 0
    # Need the fetchall due to non-async here
    for row in txn.fetchall():
        if row["id"] not in nwsli_provider:
            news += 1
            nwsli_provider[row["id"]] = row
    # Load up aliases for stations that changed IDs over the years
    txn.execute(
        "SELECT id, value from stations t JOIN station_attributes a "
        "on (t.iemid = a.iemid) WHERE attr = 'WAS'"
    )
    for row in txn.fetchall():  # Force sync
        if row["id"] not in nwsli_provider:
            LOG.info("Can't remap unknown %s", row["id"])
            continue
        nwsli_provider[row["value"]] = nwsli_provider[row["id"]]

    LOG.info("Loaded %s new stations", news)



"""XTEUS."""

import pytest

from pywwa.testing import get_example_file
from pywwa.workflows import xteus


@pytest.mark.parametrize("database", ["iem"])
def test_xteus(cursor):
    """Test XTEUS"""
    data = get_example_file("XTEUS.txt")
    res = xteus.process_data(cursor, data.replace("\003", ""))
    assert res is not None



"""Move DSM messages into the text database with the proper PIL."""

# Local
import re

import click
from pyiem.nws import product

from pywwa import common
from pywwa.database import get_dbconnc
from pywwa.ldm import bridge


def workflow(raw):
    """Go!"""
    pgconn, acursor = get_dbconnc("afos")

    data = raw.replace("\r\r\n", "z")
    tokens = re.findall("([A-Z0-9]{4} [DM]S.*?[=N]z)", data)

    nws = product.TextProduct(raw, ugc_provider={})

    sql = (
        "INSERT into products (pil, data, source, wmo, entered, bbb) "
        "values(%s,%s,%s,%s,%s,%s) "
    )
    pil3 = "DSM" if nws.wmo == "CDUS27" else "MSM"
    for token in tokens:
        sqlargs = (
            f"{pil3}{token[1:4]}",
            token.replace("z", "\n"),
            nws.source,
            nws.wmo,
            nws.valid.strftime("%Y-%m-%d %H:%M+00"),
            nws.bbb,
        )
        acursor.execute(sql, sqlargs)

    acursor.close()
    pgconn.commit()
    pgconn.close()
    return nws


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(workflow)



"""Ingest data from NWWS-OI."""

from pyiem.nws.product import TextProduct
from pyiem.util import get_properties, utc
from twisted.internet import reactor
from twisted.internet.task import LoopingCall
from twisted.names.srvconnect import SRVConnector
from twisted.words.protocols.jabber import client, error, xmlstream
from twisted.words.protocols.jabber.jid import JID
from twisted.words.xish import domish
from twisted.words.xish.xmlstream import STREAM_END_EVENT

MUC_ROOM = "nwws@conference.nwws-oi.weather.gov"


class Client:
    """A Jabber Client."""

    def __init__(self, jid, secret):
        """Constructor."""
        self.outstanding_pings = []
        self.xmlstream = None
        f = client.XMPPClientFactory(jid, secret)
        f.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, self.connected)
        f.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self.authd)
        connector = SRVConnector(
            reactor,
            "xmpp-client",
            jid.host,
            f,
            defaultPort=5222,
        )
        connector.connect()

    def connected(self, xs):
        """Connected."""
        print("Connected.")
        self.xmlstream = xs
        self.xmlstream.addObserver("/message", self.on_message)
        self.xmlstream.addObserver("/iq", self.on_iq)

    def authd(self, _xs):
        """authedn..."""
        self.outstanding_pings = []
        presence = domish.Element(("jabber:client", "presence"))
        presence["to"] = f"{MUC_ROOM}/daryl.herzmann_{utc():%Y%m%d%H%M}"
        self.xmlstream.send(presence)
        lc = LoopingCall(self.housekeeping)
        lc.start(60)
        self.xmlstream.addObserver(STREAM_END_EVENT, lambda _x: lc.stop)

    def housekeeping(self):
        """
        This gets exec'd every minute to keep up after ourselves
        1. XMPP Server Ping
        2. Update presence
        """
        if self.outstanding_pings:
            print(f"Currently unresponded pings: {self.outstanding_pings}")
        if len(self.outstanding_pings) > 5:
            self.outstanding_pings = []
            if self.xmlstream is not None:
                # Unsure of the proper code that a client should generate
                exc = error.StreamError("gone")
                self.xmlstream.sendStreamError(exc)
            return
        if self.xmlstream is None:
            print("xmlstream is None, not sending ping")
            return
        utcnow = utc()
        ping = domish.Element((None, "iq"))
        ping["to"] = "nwws-oi.weather.gov"
        ping["type"] = "get"
        pingid = f"{utcnow:%Y%m%d%H%M}"
        ping["id"] = pingid
        ping.addChild(domish.Element(("urn:xmpp:ping", "ping")))
        self.outstanding_pings.append(pingid)
        self.xmlstream.send(ping)

    def on_iq(self, elem: domish.Element):
        """Process IQ message."""
        typ = elem.getAttribute("type")
        # A response is being requested of us.
        if typ == "get" and elem.firstChildElement().name == "ping":
            # Respond to a ping request.
            pong = domish.Element((None, "iq"))
            pong["type"] = "result"
            pong["to"] = elem["from"]
            pong["from"] = elem["to"]
            pong["id"] = elem["id"]
            self.xmlstream.send(pong)
        # We are getting a response to a request we sent, maybe.
        elif typ == "result":
            if elem.getAttribute("id") in self.outstanding_pings:
                self.outstanding_pings.remove(elem.getAttribute("id"))

    def on_message(self, elem):
        """Callback."""
        if elem.hasAttribute("type") and elem["type"] == "groupchat":
            self.processMessageGC(elem)

    def processMessageGC(self, elem):
        """Got message."""
        if not elem.x:
            print(f"Unknown elem? {elem.toXml()}")
            return
        unixtext = str(elem.x)  # Unsure if this is the best way?
        noaaport = "\001" + unixtext.replace("\n\n", "\r\r\n")
        # Ensure product ends with \n
        if noaaport[-1] != "\n":
            noaaport = noaaport + "\r\r\n"
        noaaport = noaaport + "\003"
        try:
            tp = TextProduct(noaaport, parse_segments=False, ugc_provider={})
            if tp.afos is not None and tp.afos.startswith(("TOR", "SVR")):
                print(utc(), tp.get_product_id())
        except Exception as exp:
            print(f"Failed to parse: {exp}")
        # Someday perhaps fix the ldm sequence number is only 3 char
        with open(f"/mesonet/tmp/nwwsoi/{utc():%Y%m%d%H}.txt", "ab") as fh:
            fh.write(noaaport.encode("utf-8"))


def main():
    """Go Main Go."""
    props = get_properties()
    Client(
        JID(f"{props['nwws-oi.username']}@nwws-oi.weather.gov"),
        props["nwws-oi.password"],
    )
    reactor.run()


if __name__ == "__main__":
    main()



"""parse CWA"""

import pytest

from pywwa.database import get_dbconnc
from pywwa.testing import get_example_file
from pywwa.workflows import cwa


@pytest.mark.parametrize("database", ["postgis"])
def test_goodline(cursor):
    """Test goodline"""
    conn, mcursor = get_dbconnc("mesosite")
    cwa.load_database(mcursor)
    conn.close()
    prod = cwa.process_data(cursor, get_example_file("CWA.txt"))
    assert prod is not None



"""Save our local NEXRAD level 3 RADARs to staging."""

import os
import subprocess
from datetime import date, datetime, timedelta

from pyiem.util import logger

LOG = logger()
NEXRADS = "DMX DVN OAX ARX FSD MPX EAX ABR UDX".split()


def run(dt: datetime):
    """Process this date please"""
    yyyymmdd = dt.strftime("%Y%m%d")
    for nexrad in NEXRADS:
        mydir = f"/data/gempak/nexrad/NIDS/{nexrad}"
        if not os.path.isdir(mydir):
            LOG.info("creating %s", mydir)
            os.makedirs(mydir)
        os.chdir(mydir)
        cmd = (
            f"tar -czf /mesonet/tmp/{nexrad}_{yyyymmdd}.tgz "
            f"???/???_{yyyymmdd}_*"
        )
        LOG.debug(cmd)
        with subprocess.Popen(
            cmd,
            shell=True,
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        ) as proc:
            (stdout, stderr) = proc.communicate()
        if stdout != b"" or stderr != b"":
            LOG.info(
                "%s resulted in\nstdout: %s\nstderr: %s",
                cmd,
                stdout.decode("ascii", "ignore"),
                stderr.decode("ascii", "ignore"),
            )

    rpath = f"/stage/IowaNexrad3/{dt:%Y/%m}"
    cmd = (
        "rsync --remove-source-files -a --rsync-path "
        f'"mkdir -p {rpath} && rsync" '
        f"/mesonet/tmp/???_{yyyymmdd}.tgz "
        f"meteor_ldm@akrherz-desktop.agron.iastate.edu:{rpath}"
    )
    LOG.debug(cmd)
    subprocess.call(cmd, shell=True)


def main():
    """Go Main Go"""
    yesterday = date.today() - timedelta(days=1)
    run(yesterday)


if __name__ == "__main__":
    main()



"""Some magic to feed the SEL, SAW, and WWP products into the LDM bridge."""

import sys
import time


def main():
    """Go Main Go."""
    for prod in ["SEL", "SAW", "WWP"]:
        sys.stderr.write(f"Writing {prod}\n")
        with open(f"examples/{prod}.txt", "rb") as fh:
            sys.stdout.buffer.write(fh.read())
            sys.stdout.flush()
            time.sleep(10)
    sys.stderr.write("Sleeping 120\n")
    time.sleep(120)


if __name__ == "__main__":
    main()



"""Process ridge exchange messages that contain imagery."""

import json
import os
import subprocess
from datetime import datetime, timedelta, timezone

import pika


def get_rabbitmqconn():
    """Load the configuration."""
    fn = os.sep.join([os.path.dirname(__file__), "rabbitmq.json"])
    with open(fn, encoding="utf-8") as fh:
        config = json.load(fh)
    return pika.BlockingConnection(
        pika.ConnectionParameters(
            host=config["host"],
            port=config["port"],
            virtual_host=config["vhost"],
            credentials=pika.credentials.PlainCredentials(
                config["user"], config["password"]
            ),
        )
    )


def generate_image(_ch, _method, properties, body):
    """
    Generate a Lite Image given the ActiveMQ message
    @param pyactivemq.BytesMessage
    """
    siteID = properties.headers["siteID"]
    ticks = properties.headers["validTime"]
    productID = properties.headers["productID"]
    vcp = properties.headers["vcp"]
    upperLeftLat = float(properties.headers["upperLeftLat"])
    upperLeftLon = float(properties.headers["upperLeftLon"])
    lowerRightLat = float(properties.headers["lowerRightLat"])
    lowerRightLon = float(properties.headers["lowerRightLon"])
    # Convert Java ticks into local time
    gts = datetime(1970, 1, 1) + timedelta(seconds=ticks / 1000)
    gts = gts.replace(tzinfo=timezone.utc)
    utcnow = datetime.now(timezone.utc)
    routes = "ac"
    if (utcnow - gts).seconds > 3600:
        routes = "a"

    metadata = {"meta": {}}
    metadata["meta"]["valid"] = gts.strftime("%Y-%m-%dT%H:%M:%SZ")
    metadata["meta"]["product"] = productID
    metadata["meta"]["site"] = siteID
    metadata["meta"]["vcp"] = vcp
    pqstr = (
        f"gis {routes} {gts:%Y%m%d%H%M} "
        f"gis/images/4326/ridge/{siteID}/{productID}_0.png "
        f"GIS/ridge/{siteID}/{productID}/{siteID}_{productID}_"
        f"{gts:%Y%m%d%H%M}.png png"
    )

    pngfn = f"/tmp/{siteID}_{productID}_{gts:%H%M}.png"
    with open(pngfn, "wb") as fh:
        fh.write(body)

    metafn = f"/tmp/{siteID}_{productID}_{gts:%H%M}.json"
    with open(metafn, "w", encoding="ascii") as fh:
        json.dump(metadata, fh)

    wldfn = f"/tmp/{siteID}_{productID}_{gts:%H%M}.wld"
    with open(wldfn, "w", encoding="ascii") as fh:
        _val = (lowerRightLon - upperLeftLon) / 1000.0
        fh.write(f"{_val:.6f}\n")  # dx
        fh.write("0.0\n")
        fh.write("0.0\n")
        _val = 0 - (upperLeftLat - lowerRightLat) / 1000.0
        fh.write(f"{_val:.6f}\n")  # dy
        fh.write(f"{upperLeftLon:.6f}\n")  # UL Lon
        fh.write(f"{upperLeftLat:.6f}\n")  # UL Lat

    # Use -i to allow for duplicate file content as the product id *should*
    # always be unique
    pqstr = f"pqinsert -i -p '{pqstr}' {pngfn}"
    subprocess.call(pqstr, shell=True)
    subprocess.call(pqstr.replace("png", "wld"), shell=True)
    metapq = pqstr.replace("png", "json").replace(" ac ", " c ")
    subprocess.call(metapq, shell=True)

    for fn in [pngfn, wldfn, metafn]:
        if os.path.isfile(fn):
            os.unlink(fn)
        else:
            print(f"Strange file: {fn} was missing, but should be deleted")


def run():
    """Go run Go."""
    with open("ridge_processor.pid", "w", encoding="ascii") as fh:
        fh.write(f"{os.getpid()}")

    connection = get_rabbitmqconn()
    channel = connection.channel()

    # Declare the exchange
    channel.exchange_declare(
        "ridgeProductExchange", exchange_type="fanout", durable=True
    )

    # Create an exclusive, auto-delete queue for this consumer
    result = channel.queue_declare(queue="", exclusive=True)
    queue_name = result.method.queue

    # Bind the queue to the exchange
    channel.queue_bind(exchange="ridgeProductExchange", queue=queue_name)

    channel.basic_consume(queue_name, generate_image, auto_ack=True)

    channel.start_consuming()


if __name__ == "__main__":
    run()



"""For testing purposes, copy IEM network into mesosite."""

# stdlib
import sys

# Third Party
import httpx

from pywwa.database import get_dbconnc

SERVICE = "https://mesonet.agron.iastate.edu/geojson/network.py"


def main(argv):
    """Go Main Go."""
    pgconn, cursor = get_dbconnc("mesosite")
    network = argv[1]

    req = httpx.get(f"{SERVICE}?network={network}")
    jdata = req.json()
    for feat in jdata["features"]:
        site = feat["properties"]
        [lon, lat] = feat["geometry"]["coordinates"]
        if site["wfo"] is None:
            site["wfo"] = ""
        cursor.execute(
            "SELECT iemid from stations WHERE id = %s and network = %s",
            (site["sid"], network),
        )
        if cursor.rowcount == 0:
            cursor.execute(
                "INSERT into stations(id, network) values (%s, %s)",
                (site["sid"], network),
            )
        giswkt = f"SRID=4326;POINT({lon} {lat})"
        cursor.execute(
            """
            UPDATE stations SET name = %s, state = %s, elevation = %s,
            geom = %s, county = %s, wfo = %s, tzname = %s
            WHERE id = %s and network = %s
            """,
            (
                site["sname"],
                site["state"],
                site["elevation"],
                giswkt,
                site["county"],
                site["wfo"],
                site["tzname"],
                site["sid"],
                network,
            ),
        )

    cursor.close()
    pgconn.commit()


if __name__ == "__main__":
    main(sys.argv)



"""Parse CLI text products

The CLI report has lots of good data that is hard to find in other products,
so we take what data we find in this product and overwrite the database
storage of what we got from the automated observations
"""

# Local
from pywwa.workflows.cli import main

if __name__ == "__main__":
    # Do Stuff
    main()



"""
Chunk the MOS text data into easier to search values.
"""

# stdlib
import re

import click
from pyiem.nws import product

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def real_process(txn, data):
    """Go!"""
    prod = product.TextProduct(data, ugc_provider={})
    # replicate functionality in pyiem/nws/products/mos.py
    header = (
        f"000 \n{prod.wmo} {prod.source} {prod.valid:%d%H%M}\n{prod.afos}\n"
    )
    raw = prod.unixtext + "\n"
    # Since we only do realtime processing, this is OK, I hope
    sections = raw.split("\x1e")

    for sect in sections:
        tokens = re.findall(
            r"(^[A-Z0-9_]{3,10}\s+....? V?[0-9]?\.?[0-9]?\s?....? GUIDANCE)",
            sect,
        )
        if not tokens:
            continue
        # Only take 4 char IDs :/
        if len(sect[:100].split()[0]) != 4:
            continue
        # The NWS cut some corners and ended up not using proper TTAAII values
        # for non US data.
        ttaaii = prod.wmo
        if sect[0] != "K" and ttaaii[2:4] == "US":
            ttaaii = f"{ttaaii[:2]}{sect[0]}{sect[0]}{ttaaii[4:6]}"
        txn.execute(
            "INSERT into products (pil, data, source, entered, wmo) "
            "values (%s, %s, %s, %s, %s)",
            (
                prod.afos[:3] + sect[1:4],
                header + sect.replace(";;;", "\n"),
                prod.source,
                prod.valid,
                ttaaii,
            ),
        )


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(real_process, dbpool=get_database("afos"))



"""
VTEC product ingestor

The warnings table has the following timestamp based columns, this gets ugly
with watches.  Lets try to explain

    issue   <- VTEC timestamp of when this event was valid for
    expire  <- When does this VTEC product expire
    updated <- Product Timestamp of when a product gets updated
    init_expire <- When did this product initially expire
    product_issue <- When was this product issued by the NWS
"""

# Local
from pywwa.workflows.vtec import main

if __name__ == "__main__":
    main()



"""Generate a color ramp image, please."""

import sys

import netCDF4
import numpy as np
import pandas as pd
from PIL import Image, ImageDraw, ImageFont

CHANNEL_LOOKUP = {
    1: "greys",
    2: "greys",
    3: "greys",
    4: "greys",
    5: "greys",
    6: "greys",
    7: "greys",
    8: "wv",
    9: "wv",
    10: "wv",
    11: "greys",
    12: "greys",
    13: "ir",
    14: "ir",
    15: "ir",
    16: "greys",
}


def compute_range_units(channel):
    """Use example netcdf files locally."""
    nc = netCDF4.Dataset(
        f"OR_ABI-L2-CMIPC-M6C{channel:02.0f}_G16_"
        "s20193251636150_e20193251636150_c20193251636150.nc"
    )
    ncvar = nc.variables["Sectorized_CMI"]
    valid_min = ncvar.valid_min * ncvar.scale_factor + ncvar.add_offset
    valid_max = ncvar.valid_max * ncvar.scale_factor + ncvar.add_offset
    units = ncvar.units
    nc.close()
    return valid_min, valid_max, units.replace("kelvin", "K")


def main(argv):
    """Go Main Go."""
    channel = int(argv[1])
    df = pd.read_csv(
        f"{CHANNEL_LOOKUP[channel]}ramp.txt",
        names=["value", "r", "g", "b"],
        sep=" ",
    )
    vmin, vmax, units = compute_range_units(channel)
    data = np.linspace(vmin, vmax, 256)
    # We are using the netcdf2png code to figure out how values are mapped
    # into colors.
    if channel in [1, 2]:
        imgdata = np.where(
            data < 0.91,
            np.sqrt(data),
            0.9539 + (1.0 - (1.16 - data) / 0.25) * 0.0461,
        )
        imgdata = np.where(imgdata > 1, 1, imgdata) * 255
    elif channel in [3, 4, 5, 6]:
        # guidance is to take square root of data and apply grayscale
        imgdata = data**0.5 * 255
    elif channel in [7]:
        imgdata = np.where(data < 242, (418.0 - data), ((330.0 - data) * 2.0))
    elif channel in [8, 9, 10]:
        # Convert to Celsius?
        imgdata = np.digitize(data - 273.15, df["value"].values)
    elif channel in [13, 14, 15]:
        imgdata = np.where(data < 242, (418.0 - data), ((330.0 - data) * 2.0))
    else:
        # scale from min to max
        imgdata = (data - vmin) / (vmax - vmin) * 255.0

    # OK, we have data, which is our raw values and imgdata, which is how the
    # data maps to a 0-255 value.

    # Create a 256x30 image
    img = np.zeros((30, 256), dtype=np.uint8)
    img[:, :] = imgdata.astype(np.uint8)
    img[15:, :] = 0

    font = ImageFont.truetype(
        "/usr/share/fonts/liberation-mono/LiberationMono-Regular.ttf",
        10,
    )
    png = Image.fromarray(img, mode="L")
    colors = df[["r", "g", "b"]].values.ravel()
    # last index needs to be white
    colors[-3:] = 255
    png.putpalette(list(colors))
    draw = ImageDraw.Draw(png)

    # sample every 50 rows in df
    for x, row in df.iloc[50::50].iterrows():
        print(imgdata[x], data[x], row["value"])
        txt = f"{data[x]:.1f}"
        draw.line([x, 17, x, 10], fill=255)
        draw.text((x - 15, 18), txt, fill=255, font=font)

    draw.text((0, 17), units, fill=255, font=font)

    png.save(f"goes_c{channel:02.0f}.png")


if __name__ == "__main__":
    main(sys.argv)



"""Test fd_parser."""

# 3rd Party
import pytest
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import fd


@pytest.mark.parametrize("database", ["asos"])
def test_processor(cursor):
    """Test basic parsing."""
    data = get_example_file("FD1US1.txt")
    pywwa.CTX["utcnow"] = utc(2023, 3, 9, 1, 59)
    prod = fd.processor(cursor, data)
    assert prod.valid == pywwa.CTX["utcnow"]



"""Utility Functions."""

# stdlib
import os


def get_example_filepath(filename):
    """Return a full filepath."""
    return f"{os.getcwd()}/examples/{filename}"


def get_example_file(filename):
    """Return the contents of an example file."""
    fullpath = get_example_filepath(filename)
    # Need to ensure that CRCRLF remain intact
    with open(fullpath, "rb") as fh:
        return fh.read().decode("utf-8")



"""Convert the Unidata GOES product netcdf files to IEM PNG.

The devil is in the details here :)

"""

import datetime
import json
import os
import subprocess
import sys
import tempfile
import traceback

import inotify.adapters
import numpy as np
from netCDF4 import Dataset
from PIL import Image
from pyiem.util import logger, utc

LOG = logger()
ISO = "%Y-%m-%dT%H:%M:%SZ"
DIRPATH = "/mnt/goestemp"
RAMPS = {}
SECTORS = {
    "C": "CONUS",
    "HI": "Hawaii",
    "PR": "PuertoRico",
    "AK": "Alaska",
    "F": "FullDisk",
    "M1": "Mesoscale-1",
    "M2": "Mesoscale-2",
}


def build_ramps():
    """Make our color ramps."""
    # WV
    RAMPS["wv"] = np.loadtxt("wvramp.txt")
    RAMPS["ir"] = np.loadtxt("irramp.txt")
    # This is matplotlib's ramp
    RAMPS["greys"] = np.loadtxt("greysramp.txt")


def make_image(nc):
    """Generate a PIL Image.

    Thanks to help/code from College of Dupage, we have a fighting chance
    understanding how we should convert the raw netcdf data into PNG with a
    good color ramp for the data provided by the channel.  Some notes here:

    - The matplotlib Greys_r ramp is not exactly linear.

    """
    channel = nc.channel_id
    ncvar = nc.variables["Sectorized_CMI"]
    data = ncvar[:]
    valid_min = ncvar.valid_min * ncvar.scale_factor + ncvar.add_offset
    valid_max = ncvar.valid_max * ncvar.scale_factor + ncvar.add_offset

    # default grayscale
    colors = RAMPS["greys"][:, 1:].astype("i").ravel()
    if channel in [1, 2]:
        # Sometimes, values can be greater than 1, so we apply a suggested
        # conversion by Pete Pokerant, see pjpokran/goes16codes
        data = np.where(
            data < 0.91,
            np.sqrt(data),
            0.9539 + (1.0 - (1.16 - data) / 0.25) * 0.0461,
        )
        # Belt and Suspenders as values could be higher than 1.16?
        imgdata = np.where(data > 1, 1, data) * 255
    elif channel in [3, 4, 5, 6]:
        # guidance is to take square root of data and apply grayscale
        imgdata = data**0.5 * 255
    elif channel in [7]:
        imgdata = np.where(data < 242, (418.0 - data), ((330.0 - data) * 2.0))
    elif channel in [8, 9, 10]:
        # Convert to Celsius?
        imgdata = np.digitize(data - 273.15, RAMPS["wv"][:, 0])
        colors = RAMPS["wv"][:, 1:].astype("i").ravel()
    elif channel in [13, 14, 15]:
        imgdata = np.where(data < 242, (418.0 - data), ((330.0 - data) * 2.0))
        colors = RAMPS["ir"][:, 1:].astype("i").ravel()
    else:
        # scale from min to max
        imgdata = (data - valid_min) / (valid_max - valid_min) * 255.0
    png = Image.fromarray(imgdata.astype("u1"))
    png.putpalette(list(colors))
    return png


def get_sector(fn):
    """Figure out the label to associate with this file, sigh."""
    # NB nc.source_scene is not the region of view
    tokens = fn.split("-")
    s = tokens[2].replace("CMIP", "")
    return SECTORS.get(s, s)


def process(ncfn):
    """Actually process this file!"""
    if not os.path.isfile(ncfn):
        return
    sector = get_sector(ncfn)
    with Dataset(ncfn) as nc:
        valid = datetime.datetime.strptime(nc.start_date_time, "%Y%j%H%M%S")
        bird = nc.satellite_id
        channel = nc.channel_id
        lon_0 = nc.satellite_longitude
        h = nc.variables["fixedgrid_projection"].perspective_point_height
        swa = nc.variables["fixedgrid_projection"].sweep_angle_axis
        dx = nc.variables["x"].scale_factor / 1e6 * h
        y0 = nc.variables["y"][0] / 1e6 * h
        x0 = nc.variables["x"][0] / 1e6 * h
        dy = nc.variables["y"].scale_factor / 1e6 * h
        png = make_image(nc)

    tmpfd = tempfile.NamedTemporaryFile(delete=False)
    png.save(tmpfd, format="PNG")
    # close/destroy the PIL object to hopefully stop memory leaks
    png.close()
    del png
    tmpfd.close()

    pqstr = (
        f"gis c {valid:%Y%m%d%H%M} gis/images/GOES/{sector.lower()}/"
        f"channel{channel:02.0f}/{bird}_C{channel:02.0f}.png "
        f"{valid:%Y%m%d%H%M%S} png"
    )
    subprocess.call(["pqinsert", "-i", "-p", pqstr, tmpfd.name])
    os.unlink(tmpfd.name)
    tmpfd = tempfile.NamedTemporaryFile(mode="w", delete=False)
    args = [dx, 0, 0, dy, x0, y0]
    tmpfd.write("\n".join([str(s) for s in args]))
    tmpfd.close()
    subprocess.call(
        f"pqinsert -i -p '{pqstr.replace('png', 'wld')}' {tmpfd.name}",
        shell=True,
    )
    with open(tmpfd.name, "w", encoding="utf8") as fh:
        fh.write(
            f"PROJECTION\nproj=geos\nh={h}\nlon_0={lon_0}\nsweep={swa}\nEND\n"
        )
    subprocess.call(
        f"pqinsert -i -p '{pqstr.replace('png', 'msinc')}' {tmpfd.name}",
        shell=True,
    )

    # Make the json metadata file
    meta = {
        "generated_at": utc().strftime(ISO),
        "meta": {
            "valid": valid.strftime(ISO),
            "proj4str": f"+proj=geos +h={h} +lon_0={lon_0} +sweep={swa}",
        },
    }
    with open(tmpfd.name, "w", encoding="utf8") as fh:
        json.dump(meta, fh)
    subprocess.call(
        f"pqinsert -i -p '{pqstr.replace('png', 'json')}' {tmpfd.name}",
        shell=True,
    )

    os.unlink(tmpfd.name)


def main(argv):
    """Go Main Go."""
    # Shard by bird, either G16 or G17
    bird = f"_{argv[1]}_"
    inotif = inotify.adapters.Inotify()
    inotif.add_watch(DIRPATH)
    try:
        for event in inotif.event_gen():
            if event is None:
                continue
            (_header, type_names, watch_path, fn) = event
            # Careful here, it seems like we get the IN_CLOSE_NOWRITE when
            # we double back with the netcdf read statement above closing.
            if "IN_CLOSE_WRITE" not in type_names:
                continue
            if not fn.endswith(".nc") or fn.find(bird) == -1:
                continue
            ncfn = f"{watch_path}/{fn}"
            try:
                process(ncfn)
            except Exception as exp:
                # Full disk will cause grief
                try:
                    with open(f"{ncfn}.error", "w", encoding="utf8") as fp:
                        fp.write(str(exp) + "\n")
                        traceback.print_exc(file=fp)
                except Exception as exp2:
                    LOG.exception(exp2)
            try:
                if os.path.isfile(ncfn):
                    os.unlink(ncfn)
            except Exception as exp:
                LOG.exception(exp)
    finally:
        inotif.remove_watch(DIRPATH)


if __name__ == "__main__":
    build_ramps()
    main(sys.argv)



"""A memcache client that is meant to be run from a thread within twisted."""

import threading

from pymemcache.client import Client


def write_memcache(key, value, expire=600):
    """Write a key/value pair to memcache"""

    def _write():
        """Do the actual write, from a thread."""
        # NB 1 second was too tight
        mc = Client(("iem-memcached", 11211), connect_timeout=6)
        mc.set(key, value, expire=expire)
        mc.close()

    threading.Thread(target=_write).start()



"""Test the NLDN workflow."""

import pytest

from pywwa.testing import get_example_filepath
from pywwa.workflows.nldn import process_data


def test_nldn_empty():
    """Exercise an empty NLDN file."""
    assert process_data(None, "") is None


@pytest.mark.parametrize("database", ["nldn"])
def test_nldn_workflow(cursor):
    """Test the NLDN workflow only inserts data once, le sigh."""
    fn = get_example_filepath("nldn")
    with open(fn) as fh:
        # NB, this is verbatim from IDD, which has the NLDN prefix
        data = fh.read()
    for line in data.splitlines():
        process_data(cursor, line)
    cursor.execute(
        "select count(*) from nldn2025_07 where valid > '2025-07-23' and "
        "valid < '2025-07-24'",
    )
    assert cursor.fetchone()["count"] == 1



"""XMPP/Jabber Client Interface and Support."""

# stdlib
import inspect
import os
import re

# Third Party
import pyiem
import treq
from pyiem.util import utc
from twisted.internet import reactor as theReactor
from twisted.internet.defer import inlineCallbacks
from twisted.web import client as webclient
from twisted.words.protocols.jabber import client as jclient
from twisted.words.protocols.jabber import xmlstream
from twisted.words.protocols.jabber.jid import JID
from twisted.words.xish import domish, xpath

# Local
from pywwa import CTX, LOG, SETTINGS, shutdown

# http://stackoverflow.com/questions/7016602
webclient._HTTP11ClientFactory.noisy = False  # skipcq: PYL-W0212
# create a regular expression that matches any illegal XML character
# http://stackoverflow.com/questions/1707890
ILLEGAL_XML_CHARS_RE = re.compile(
    "[\x00-\x08\x0b\x0c\x0e-\x1f\ud800-\udfff\ufffe\uffff]", re.UNICODE
)


def make_jabber_client(resource_prefix=None):
    """Generate a jabber client, please"""
    if CTX["disable_xmpp"]:
        LOG.info("XMPP disabled via command line.")
        CTX["JABBER"] = NOOPXMPP()
        return CTX["JABBER"]

    if resource_prefix is None:
        # Inspect the calling stack to determine the script name that is
        # calling us, so we can use that as the resource prefix
        fs = inspect.stack()
        resource_prefix = os.path.basename(fs[-1].filename).removesuffix(".py")

    myjid = JID(
        f"{SETTINGS.get('pywwa_jabber_username', 'iembot_ingest')}@"
        f"{SETTINGS.get('pywwa_jabber_domain', 'localhost')}/"
        f"{resource_prefix}_{utc():%Y%m%d%H%M%S}"
    )
    CTX["JABBER"] = JabberClient(
        theReactor,
        myjid,
        SETTINGS.get("pywwa_jabber_password", "secret"),
    )
    return CTX["JABBER"]


def message_processor(stanza):
    """Process a message stanza"""
    body = xpath.queryForString("/message/body", stanza)
    LOG.info("Message from %s Body: %s", stanza["from"], body)
    if body is not None and body.lower().strip() == "shutdown":
        LOG.info("I got shutdown message, shutting down...")
        shutdown()


def raw_data_in(data):
    """
    Debug method
    @param data string of what was received from the server
    """
    if isinstance(data, bytes):
        data = data.decode("utf-8", errors="ignore")
    LOG.info("RECV %s", data)


def debug(mixed):
    """
    Debug method
    @param elem twisted.works.xish
    """
    if isinstance(mixed, domish.Element):
        LOG.info(mixed.toXml().encode("utf-8"))
    else:
        LOG.info(mixed)


def raw_data_out(data):
    """
    Debug method
    @param data string of what data was sent
    """
    if isinstance(data, bytes):
        data = data.decode("utf-8", errors="ignore")
    if data == " ":
        return
    LOG.info("SEND %s", data)


class NOOPXMPP:
    """A no-operation Jabber client."""

    def keepalive(self):
        """Do Nothing."""

    def send_message(self, *args):
        """Do Nothing."""


class JabberClient:
    """
    I am an important class of a jabber client against the chat server,
    used by pretty much every ingestor as that is how messages are routed
    to nwsbot
    """

    def __init__(self, reactor, myjid: JID, secret):
        """
        Constructor

        @param jid twisted.words.jid object
        """
        self.reactor = reactor
        self.myjid = myjid
        self.xmlstream = None
        self.authenticated = False
        self.routerjid = (
            f"{SETTINGS.get('bot.username', 'iembot')}@"
            f"{SETTINGS.get('pywwa_jabber_domain', 'localhost')}"
        )
        f = jclient.XMPPClientFactory(myjid, secret)
        # Important, we need to capture the xmlstream so to handle errors
        f.addBootstrap(xmlstream.STREAM_CONNECTED_EVENT, self.connected)
        # Send initial presence so that we are `online`
        f.addBootstrap(xmlstream.STREAM_AUTHD_EVENT, self.authd)
        f.addBootstrap(xmlstream.STREAM_END_EVENT, self.disconnect)
        f.addBootstrap(xmlstream.INIT_FAILED_EVENT, self.init_failed)

        # It is difficult to hardcode the SRV record, so we use old style
        xmpp_host = SETTINGS.get("pywwa_jabber_host", myjid.host)
        xmpp_port = int(SETTINGS.get("pywwa_jabber_port", 5222))
        reactor.connectTCP(xmpp_host, xmpp_port, f)

    def init_failed(self, failure):
        """
        Called when we fail to initialize the connection
        @param failure twisted.python.failure.Failure
        """
        LOG.error("Failed to initialize XMPP connection: %s", failure)
        if self.xmlstream is not None:
            self.xmlstream.sendFooter()

    def connected(self, xstream: xmlstream.XmlStream):
        """
        Callbacked once we are connected to the server
        @param xs twisted.words.xish.xmlstream
        """
        LOG.info("Connected to %s", self.myjid.host)
        self.xmlstream = xstream
        self.xmlstream.rawDataInFn = raw_data_in
        self.xmlstream.rawDataOutFn = raw_data_out
        # Process Messages
        self.xmlstream.addObserver("/message/body", message_processor)
        # Process IQ Requests
        self.xmlstream.addObserver("/iq", self.iq_processor)

    def authd(self, _xstream):
        """
        Callbacked once authentication succeeds
        @param xs twisted.words.xish.xmlstream
        """
        LOG.info("Logged in as %s", self.myjid)
        self.authenticated = True

        # Send initial presence
        presence = domish.Element(("jabber:client", "presence"))
        presence.addElement("status").addContent("Online")
        self.xmlstream.send(presence)

    def iq_processor(self, iq: domish.Element):
        """Process an IQ request"""
        child_el = iq.firstChildElement()
        # This is fine, not all IQs have children
        if child_el is None:
            return
        # Handle pings
        if child_el.name == "ping":
            pong = domish.Element((None, "iq"))
            pong["type"] = "result"
            pong["to"] = iq["from"]
            pong["from"] = iq["to"]
            pong["id"] = iq["id"]
            self.xmlstream.send(pong)
            return
        # Handle <query xmlns="jabber:iq:version"/>
        if child_el.name == "query" and child_el.uri == "jabber:iq:version":
            query = domish.Element((None, "iq"))
            query["type"] = "result"
            query["to"] = iq["from"]
            query["from"] = iq["to"]
            query["id"] = iq["id"]
            query.addElement("query", "jabber:iq:version")
            query.firstChildElement().addElement("name").addContent(
                self.myjid.resource
            )
            query.firstChildElement().addElement("version").addContent(
                pyiem.__version__
            )
            self.xmlstream.send(query)
            return

    def disconnect(self, _xs):
        """
        Called when we are disconnected from the server, I guess
        """
        LOG.info("SETTING authenticated to false!")
        self.xmlstream = None
        self.authenticated = False

    @inlineCallbacks
    def send_message(self, body, html, xtra):
        """
        Send a message to nwsbot.  This message should have
        @param body plain text variant
        @param html html version of the message
        @param xtra dictionary of stuff that tags along
        """
        if not self.authenticated:
            LOG.info("No Connection, Lets wait and try later...")
            self.reactor.callLater(
                3,
                self.send_message,
                body,
                html,
                xtra,  # @UndefinedVariable
            )
            return
        message = domish.Element(("jabber:client", "message"))
        message["to"] = self.routerjid
        message["type"] = "chat"

        body = ILLEGAL_XML_CHARS_RE.sub("", body)
        if html:
            html = ILLEGAL_XML_CHARS_RE.sub("", html)
        message.addElement("body", None, body)
        helem = message.addElement(
            "html", "http://jabber.org/protocol/xhtml-im"
        )
        belem = helem.addElement("body", "http://www.w3.org/1999/xhtml")
        belem.addRawXml(html or body)
        # channels is of most important
        xelem = message.addElement("x", "nwschat:nwsbot")
        for key in xtra.keys():
            if isinstance(xtra[key], list):
                xelem[key] = ",".join(xtra[key])
            else:
                xelem[key] = xtra[key]

        # If we have twitter_media, we want to attempt to get the content
        # cached before iembot asks for it a million times
        url = xtra.get("twitter_media")
        if url is not None:
            try:
                # https leaks memory twisted/treq/issues/380
                _r = yield treq.get(url.replace("https", "http"), timeout=120)
            except Exception as exp:
                LOG.info("twitter_media request for %s failed: %s", url, exp)
        self.reactor.callFromThread(self.xmlstream.send, message)



"""Test cf6_parser."""

# 3rd Party
import pytest
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import cf6


@pytest.mark.parametrize("database", ["iem"])
def test_processor(cursor):
    """Test basic parsing."""
    data = get_example_file("CF6.txt")
    pywwa.CTX["utcnow"] = utc(2020, 11, 25, 9, 20)
    prod = cf6.processor(cursor, data)
    assert prod.valid == pywwa.CTX["utcnow"]



"""Test vtec_parser."""

from pywwa.database import get_database
from pywwa.testing import get_example_file
from pywwa.workflows import vtec


def test_parser():
    """Test that we can parse!"""
    vtec.process_data({}, get_database("postgis"), get_example_file("TOR.txt"))



"""SPS product ingestor"""

from functools import partial

# 3rd Party
import click
from pyiem.database import get_sqlalchemy_conn
from pyiem.nws.products.sps import parser
from pyiem.nws.ugc import UGCProvider

# Local
from pywwa import LOG, common
from pywwa.database import get_database, load_nwsli
from pywwa.ldm import bridge

NWSLI_DICT = {}


def real_process(ugc_dict, txn, raw):
    """Really process!"""
    if raw.find("$$") == -1:
        LOG.info("$$ was missing from this product")
        raw += "\r\r\n$$\r\r\n"
    prod = parser(
        raw,
        utcnow=common.utcnow(),
        ugc_provider=ugc_dict,
        nwsli_provider=NWSLI_DICT,
    )
    if common.dbwrite_enabled():
        prod.sql(txn)
    baseurl = common.SETTINGS.get("pywwa_product_url", "pywwa_product_url")
    jmsgs = prod.get_jabbers(baseurl)
    for mess, htmlmess, xtra in jmsgs:
        common.send_message(mess, htmlmess, xtra)
    if prod.warnings:
        common.email_error("\n\n".join(prod.warnings), prod.text)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    with get_sqlalchemy_conn("postgis") as conn:
        ugc_dict = UGCProvider(pgconn=conn)
    load_nwsli(NWSLI_DICT)
    func = partial(real_process, ugc_dict)
    bridge(func, dbpool=get_database("postgis"))



"""Test pywwa/__init__.py"""

import os
import shutil

import pytest

import pywwa


def test_invalid_filename():
    """Test that exception is raised"""
    with pytest.raises(FileNotFoundError):
        pywwa.get_data_filepath("foo.txt")


def test_load_settings():
    """Test that we can load settings."""
    if os.path.isfile(pywwa.SETTINGS_FN):
        shutil.move(pywwa.SETTINGS_FN, f"{pywwa.SETTINGS_FN}.save")
    pywwa.load_settings()
    assert isinstance(pywwa.SETTINGS, dict)
    if os.path.isfile(f"{pywwa.SETTINGS_FN}.save"):
        shutil.move(f"{pywwa.SETTINGS_FN}.save", pywwa.SETTINGS_FN)



"""
Support SPC's MCD product
Support WPC's FFG product
"""

# 3rd Party
import click
from pyiem.nws.products.mcd import parser as mcdparser

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def find_cwsus(txn, prod):
    """
    Provided a database transaction, go look for CWSUs that
    overlap the discussion geometry.
    ST_Overlaps do the geometries overlap
    ST_Covers does polygon exist inside CWSU
    """
    wkt = f"SRID=4326;{prod.geometry.wkt}"
    txn.execute(
        "select distinct id from cwsu WHERE st_overlaps(%s, geom) or "
        "st_covers(geom, %s) ORDER by id ASC",
        (wkt, wkt),
    )
    cwsus = []
    for row in txn.fetchall():
        cwsus.append(row["id"])
    return cwsus


def real_process(txn, raw):
    """ "
    Actually process a single MCD
    """
    prod = mcdparser(raw.upper(), ugc_provider={})
    prod.cwsus = find_cwsus(txn, prod)

    jmsgs = prod.get_jabbers(
        common.SETTINGS.get("pywwa_product_url", "pywwa_product_url")
    )
    for j in jmsgs:
        common.send_message(j[0], j[1], j[2])
    if common.dbwrite_enabled():
        prod.database_save(txn)
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), raw)
    return prod


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(real_process, dbpool=get_database("postgis"))



"""
Compare our PIREPs data against what is at aviation wx JSON service
"""

import datetime

import httpx
from pyiem.util import logger

from pywwa.database import get_dbconn


def main():
    """Go Main Go."""
    log = logger()
    pgconn = get_dbconn("postgis")
    cursor = pgconn.cursor()

    avwx = httpx.get("http://aviationweather.gov/gis/scripts/AirepJSON.php")
    avwx = avwx.json()

    mine = {}
    cursor.execute(
        """
        SELECT valid at time zone 'UTC', ST_x(geom::geometry),
        ST_y(geom::geometry), report
        from pireps WHERE valid > (now() - '60 minutes'::interval)
    """
    )
    for row in cursor:
        key = "/".join(row[3].replace(" ", "").split("/")[:3])
        mine[key] = row

    floor = None
    for feature in avwx["features"]:
        if feature["properties"]["airepType"] != "PIREP":
            continue
        ts = datetime.datetime.strptime(
            feature["properties"]["obsTime"], "%Y-%m-%dT%H:%M:%SZ"
        )
        if floor is None:
            floor = ts
        lon, lat = feature["geometry"]["coordinates"]
        key = "/".join(
            feature["properties"]["rawOb"].replace(" ", "").split("/")[:3]
        )
        if key not in mine:
            log.info("IEM  MISS %s %s", ts, feature["properties"]["rawOb"])
        else:
            error = (
                (mine[key][1] - lon) ** 2 + (mine[key][2] - lat) ** 2
            ) ** 0.5
            if error > 0.1:
                location = "/".join(
                    feature["properties"]["rawOb"].split("/")[:2]
                )
                log.info(
                    "ERROR: %5.2f IEM: %8.3f %6.3f AWX: %7.2f %5.2f %s",
                    error,
                    mine[key][1],
                    mine[key][2],
                    lon,
                    lat,
                    location,
                )
            del mine[key]

    for _, item in mine.items():
        if item[0] < floor:
            continue
        log.info("AVWX MISS %s %s", item[0], item[3])


if __name__ == "__main__":
    main()



"""Test nexrad3_attr."""

# 3rd Party
import pytest

from pywwa.testing import get_example_filepath

# Local
from pywwa.workflows import nexrad3_attr


@pytest.mark.parametrize("database", ["mesosite"])
def test_load_station_table(cursor):
    """Test loading of the station table."""
    nexrad3_attr.load_station_table(cursor)


@pytest.mark.parametrize("database", ["radar"])
def test_process(cursor):
    """Test the processing of a level III file."""
    with open(get_example_filepath("NCR_20121127_1413"), "rb") as fh:
        ctx = nexrad3_attr.process(fh)
    assert ctx["nexrad"] == "JAX"
    processed = nexrad3_attr.really_process(cursor, ctx)
    assert processed == 3


@pytest.mark.parametrize("database", ["radar"])
def test_210910_badvil(cursor):
    """Test that a missing VIL does not cause issues."""
    with open(get_example_filepath("NCR_20210911_0023"), "rb") as fh:
        ctx = nexrad3_attr.process(fh)
    assert ctx["nexrad"] == "LAS"
    processed = nexrad3_attr.really_process(cursor, ctx)
    assert processed == 2



"""PIREP parser!"""

from datetime import datetime, timedelta, timezone

import click
from pyiem.nws.products.pirep import parser as pirepparser
from twisted.internet import reactor

# Local
from pywwa import LOG, common, get_data_filepath
from pywwa.database import get_database, get_dbconnc
from pywwa.ldm import bridge

PIREPS = {}
# Load LOCS table
LOCS = {}


def cleandb():
    """To keep LSRDB from growing too big, we clean it out
    Lets hold 1 days of data!
    """
    thres = datetime.now(timezone.utc) - timedelta(hours=24 * 1)
    init_size = len(PIREPS.keys())
    # loop safety
    for key in list(PIREPS):
        if PIREPS[key] < thres:
            PIREPS.pop(key)

    fin_size = len(PIREPS.keys())
    LOG.info("cleandb() init_size: %s final_size: %s", init_size, fin_size)

    # Call Again in 30 minutes
    reactor.callLater(60 * 30, cleandb)  # @UndefinedVariable


def load_locs(txn):
    """Build locations table"""
    LOG.info("load_locs() called...")
    txn.execute(
        "SELECT id, name, st_x(geom) as lon, st_y(geom) as lat "
        "from stations WHERE network ~* 'ASOS'"
    )
    for row in txn.fetchall():
        LOCS[row["id"]] = {
            "id": row["id"],
            "name": row["name"],
            "lon": row["lon"],
            "lat": row["lat"],
        }

    fn = get_data_filepath("faa_apt.tbl")
    with open(fn, encoding="utf-8") as fh:
        for line in fh:
            if len(line) < 70 or line[0] == "!":
                continue
            sid = line[:4].strip()
            lat = float(line[56:60]) / 100.0
            lon = float(line[61:67]) / 100.0
            name = line[16:47].strip()
            if sid not in LOCS:
                LOCS[sid] = {"lat": lat, "lon": lon, "name": name}

    fn = get_data_filepath("vors.tbl")
    with open(fn, encoding="utf-8") as fh:
        for line in fh:
            if len(line) < 70 or line[0] == "!":
                continue
            sid = line[:3]
            lat = float(line[56:60]) / 100.0
            lon = float(line[61:67]) / 100.0
            name = line[16:47].strip()
            if sid not in LOCS:
                LOCS[sid] = {"lat": lat, "lon": lon, "name": name}

    # Finally, GEMPAK!
    fn = get_data_filepath("pirep_navaids.tbl")
    with open(fn, encoding="utf-8") as fh:
        for line in fh:
            if len(line) < 60 or line[0] in ["!", "#"]:
                continue
            sid = line[:4].strip()
            lat = float(line[56:60]) / 100.0
            lon = float(line[61:67]) / 100.0
            if sid not in LOCS:
                LOCS[sid] = {"lat": lat, "lon": lon}

    LOG.info("... %s locations loaded", len(LOCS))


def real_parser(txn, buf):
    """
    I'm gonna do the heavy lifting here
    """
    prod = pirepparser(
        buf, utcnow=common.utcnow(), nwsli_provider=LOCS, ugc_provider={}
    )
    prod.assign_cwsu(txn)
    for report in prod.reports:
        if report.text in PIREPS:
            report.is_duplicate = True
        PIREPS[report.text] = datetime.now(timezone.utc)

    j = prod.get_jabbers("unused")
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), buf)
    for msg in j:
        common.send_message(msg[0], msg[1], msg[2])
    if common.dbwrite_enabled():
        prod.sql(txn)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """GO Main Go."""
    conn, cursor = get_dbconnc("mesosite")
    load_locs(cursor)
    conn.close()
    bridge(real_parser, dbpool=get_database("postgis"))



"""Test the LSR parser."""

from pywwa.workflows import lsr


def test_pickling():
    """Test the lifecycle of pickling the LSRDB."""
    lsr.pickledb()
    lsr.loaddb()
    # This writes, so be careful
    lsr.cleandb()



"""AFOS Database Workflow."""

import click
from pyiem.wmo import WMOProduct
from twisted.internet import reactor

from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge
from pywwa.memclient import write_memcache

MEMCACHE_EXCLUDE = [
    "RR1",
    "RR2",
    "RR3",
    "RR4",
    "RR5",
    "RR6",
    "RR7",
    "RR8",
    "RR9",
    "ROB",
    "HML",
]


def write2memcache(product_id: str, text: str):
    """write our TextProduct to memcached"""
    # 10 minutes should be enough time
    LOG.debug("writing %s to memcache", product_id)
    write_memcache(
        product_id.encode("utf-8"),
        text.encode("utf-8"),
        expire=600,
    )


def process_data(txn, buf):
    """Actually do something with the buffer, please"""
    if buf.strip() == "":
        return None
    utcnow = common.utcnow()

    nws = WMOProduct(buf, utcnow=utcnow)

    # When we are in realtime processing, do not consider old data, typically
    # when a WFO fails to update the date in their MND
    if not common.replace_enabled():
        delta = (nws.valid - utcnow).total_seconds()
        if delta < (-180 * 86400):  # 180 days
            raise ValueError(f"Very Latent Product! {nws.valid}")
        if delta > (6 * 3600):  # Six Hours
            raise ValueError(f"Product from the future! {nws.valid}")
    if nws.warnings:
        common.email_error("\n".join(nws.warnings), buf)
    if nws.afos is None:
        if nws.source[0] not in ["K", "P"]:
            return None
        raise ValueError("TextProduct.afos is null")

    if common.replace_enabled():
        txn.execute(
            "DELETE from products where pil = %s and source = %s and "
            "entered = %s and bbb is not distinct from %s",
            (
                nws.afos,
                nws.source,
                nws.valid,
                nws.bbb,
            ),
        )
        LOG.info("Removed %s rows for %s", txn.rowcount, nws.get_product_id())

    txn.execute(
        "INSERT into products (pil, data, entered, "
        "source, wmo, bbb) VALUES(%s, %s, %s, %s, %s, %s)",
        (
            nws.afos.strip(),
            nws.unixtext,
            nws.valid,
            nws.source,
            nws.wmo,
            nws.bbb,
        ),
    )
    if nws.afos[:3] in MEMCACHE_EXCLUDE:
        return None
    # We are on a thread, so we need to send this back to the main thread
    reactor.callFromThread(write2memcache, nws.get_product_id(), nws.unixtext)
    return nws


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Fire up our workflow."""
    bridge(process_data, dbpool=get_database("afos", cp_max=5))



"""HML parser!"""

# 3rd Party
import click
from pyiem.nws.products.hml import parser as hmlparser

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def real_parser(txn, buf):
    """I'm gonna do the heavy lifting here"""
    prod = hmlparser(buf, ugc_provider={})
    if common.dbwrite_enabled():
        prod.sql(txn)
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), buf)


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(real_parser, dbpool=get_database("hml"))



"""Hit the NWSChat quasi web service to get the text IEM may have missed :("""

import sys
import tempfile

import httpx


def wrap(data):
    """convert data into more noaaportish"""
    data = data.replace("&gt;", ">").replace("&lt;", "<")
    return "\001" + data.replace("&amp;", "&").replace("\n", "\r\r\n") + "\003"


def process(j):
    """Process the json data j"""
    if not j["data"]:
        print("ERROR: No results found!")
        return

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmpfd:
        tmpfd.write(wrap(j["data"][0]["report"]))
        for svs in j["data"][0]["svs"]:
            tmpfd.write(wrap(svs))
        print(f"Created {tmpfd.name}")


def main(argv):
    """Do Work please"""
    year = argv[1]
    wfo = argv[2]
    phenomena = argv[3]
    significance = argv[4]
    eventid = argv[5]
    uri = (
        "https://nwschat.weather.gov/vtec/json-text.php?"
        "year=%s&wfo=%s&phenomena=%s&eventid=%s&significance=%s"
    ) % (year, wfo, phenomena, eventid, significance)

    req = httpx.get(uri)
    process(req.json())


if __name__ == "__main__":
    main(sys.argv)



"""Utility script to generate the HTML used for IEMBot Channel Documentation"""

import os
import re
import sys

from pyiem.nws.nwsli import NWSLI
from pyiem.nws.products import parser as productparser
from pyiem.nws.products.cwa import parser as cwaparser
from pyiem.nws.products.vtec import parser as vtec_parser
from pyiem.nws.ugc import UGC
from pyiem.reference import prodDefinitions
from pyiem.util import logger

from pywwa.database import get_dbconnc

PARSERS = {
    "CWA": cwaparser,
}
CHANNELSFN = "/home/akrherz/projects/iem/htdocs/projects/iembot/channels.html"
LOG = logger()
ugc_dict = {}
nwsli_dict = {}

C1 = "&lt;wfo&gt;"
C2 = "&lt;vtec_phenomena&gt;.&lt;vtec_significance&gt;"
C3 = "&lt;afos_pil&gt;"
C3p = "&lt;afos_pil_prefix&gt;..."
C35 = "&lt;afos_pil_first5&gt;..."
C3S = "&lt;afos_pil_first3&gt;.&lt;state&gt;"
C4 = "&lt;vtec_phenomena&gt;.&lt;vtec_significance&gt;.&lt;wfo&gt;"
C5 = "&lt;vtec_phenomena&gt;.&lt;vtec_significance&gt;.&lt;ugc&gt;"
C5s = "&lt;vtec_phenomena&gt;.&lt;vtec_significance&gt;.&lt;state&gt;"
C6 = "&lt;ugc&gt;"
C7 = "&lt;afos_pil&gt;.&lt;wfo&gt;"
C8 = "&lt;wmo_source&gt;.&lt;aaa&gt;"
C9 = "&lt;afos_pil&gt;.&lt;ugc&gt;"
D = {
    "10-313": "https://weather.gov/directives/sym/pd01003013curr.pdf",
    "10-314": "https://weather.gov/directives/sym/pd01003014curr.pdf",
    "10-315": "https://weather.gov/directives/sym/pd01003015curr.pdf",
    "10-320": "https://weather.gov/directives/sym/pd01003020curr.pdf",
    "10-330": "https://weather.gov/directives/sym/pd01003030curr.pdf",
    "10-401": "https://weather.gov/directives/sym/pd01004001curr.pdf",
    "10-511": "https://weather.gov/directives/sym/pd01005011curr.pdf",
    "10-512": "https://weather.gov/directives/sym/pd01005012curr.pdf",
    "10-513": "https://weather.gov/directives/sym/pd01005013curr.pdf",
    "10-515": "https://weather.gov/directives/sym/pd01005015curr.pdf",
    "10-517": "https://weather.gov/directives/sym/pd01005017curr.pdf",
    "10-601": "https://weather.gov/directives/sym/pd01006001curr.pdf",
    "10-803": "https://weather.gov/directives/sym/pd01008003curr.pdf",
    "10-912": "https://weather.gov/directives/sym/pd01009012curr.pdf",
    "10-922": "https://weather.gov/directives/sym/pd01009022curr.pdf",
    "10-930": "https://weather.gov/directives/sym/pd01009030curr.pdf",
    "10-1004": "https://weather.gov/directives/sym/pd01010004curr.pdf",
    "10-1701": "https://weather.gov/directives/sym/pd01017001curr.pdf",
}
SPECIAL = {
    "PTSDY1": "Storm Prediction Center Day 1 Convective Outlook",
    "PTSDY2": "Storm Prediction Center Day 2 Convective Outlook",
    "PFWFD1": "Storm Prediction Center Day 1 Fire Weather Outlook",
    "PFWFD2": "Storm Prediction Center Day 2 Fire Weather Outlook",
    "RBG94E": "Weather Prediction Center Day 1 Excessive Rainfall Outlook",
    "RBG98E": "Weather Prediction Center Day 2 Excessive Rainfall Outlook",
    "RBG99E": "Weather Prediction Center Day 3 Excessive Rainfall Outlook",
}

# TODO: TCV TSU ADR CDW DSA EQW HMW HPA LEw NUW RHW VOW PQS
# Our dictionary of products!
S1 = [C1, C2, C3, C3p, C4, C5, C5s, C6]
VTEC_PRODUCTS = [
    dict(afos="CFW", directive="10-320", channels=S1),
    dict(afos="EWW", directive="10-601", channels=S1),
    dict(afos="FFA", directive="10-922", channels=S1),
    dict(afos="FFS", directive="10-922", channels=S1),
    dict(afos="FFW", directive="10-922", channels=S1),
    dict(afos="FLS", directive="10-922", channels=S1),
    dict(afos="FLW", directive="10-922", channels=S1),
    dict(
        afos="MWW",
        directive="10-315",
        channels=[C2, C3, C3p, C4, C5, C6],
        notes=(
            "This product does not get routed to the "
            '<span class="badge bg-light text-dark">&lt;wfo&gt;</span> '
            "channel.  This is because the product is very frequently "
            "issued for offices with marine zones."
        ),
    ),
    dict(afos="NPW", directive="10-515", channels=S1),
    dict(afos="RFW", directive="10-401", channels=S1),
    dict(afos="SMW", directive="10-313", channels=S1),
    dict(afos="SQW", directive="10-511", channels=S1),
    dict(afos="SVR", directive="10-511", channels=S1),
    dict(afos="SVS", directive="10-511", channels=S1),
    dict(afos="TOR", directive="10-511", channels=S1),
    dict(afos="WCN", directive="10-511", channels=S1),
    dict(afos="WSW", directive="10-513", channels=S1),
]
S21 = [C3, C3p, C35]
S2 = [C3, C3p]
GEN_PRODUCTS = [
    dict(afos="ADA", directive="10-1701", channels=S2),
    dict(afos="ADM", directive="10-1701", channels=S2),
    dict(afos="AFD", directive="10-1701", channels=S2),
    dict(afos="AQI", directive="10-1701", channels=S2),
    dict(afos="AWU", directive="10-1701", channels=S2),
    dict(afos="AWW", directive="10-1701", channels=S2),
    dict(afos="AVA", directive="10-1701", channels=S2),
    dict(afos="AVG", directive="10-1701", channels=S2),
    dict(afos="AVW", directive="10-1701", channels=S2),
    dict(afos="AQA", directive="10-1701", channels=S2),
    dict(afos="BLU", directive="10-1701", channels=S2),
    dict(afos="CAE", directive="10-1701", channels=S2),
    dict(afos="CEM", directive="10-1701", channels=S2),
    dict(afos="CF6", directive="10-1004", channels=S2),
    dict(afos="CGR", directive="10-1701", channels=S2),
    dict(afos="CLI", directive="10-1004", channels=S2),
    dict(afos="CLM", directive="10-1004", channels=S2),
    dict(afos="CRF", directive="10-912", channels=S2),
    dict(afos="CWA", directive="10-803", channels=S2),
    dict(afos="CWF", directive="10-1701", channels=S2),
    dict(afos="DGT", directive="10-1701", channels=S2),
    dict(afos="ESF", directive="10-1701", channels=S2),
    dict(afos="EQI", directive="10-1701", channels=S2),
    dict(afos="EQR", directive="10-1701", channels=S2),
    dict(afos="EVI", directive="10-1701", channels=S2),
    dict(afos="FRW", directive="10-1701", channels=S2),
    dict(afos="FTM", directive="10-1701", channels=S2),
    dict(afos="FWA", directive="10-1701", channels=S2),
    dict(afos="FWF", directive="10-1701", channels=S2),
    dict(afos="FWS", directive="10-1701", channels=S2),
    dict(afos="GLF", directive="10-1701", channels=S2),
    dict(afos="HLS", directive="10-601", channels=S2),
    dict(afos="HCM", directive="10-1701", channels=S2),
    dict(afos="HMD", directive="10-1701", channels=S2),
    dict(afos="HWO", directive="10-517", channels=S2),
    dict(afos="HYD", directive="10-1701", channels=S2),
    dict(afos="ICE", directive="10-330", channels=S2),
    dict(afos="LAE", directive="10-1701", channels=S2),
    dict(afos="LCO", directive="10-1701", channels=S2),
    dict(
        afos="LSR",
        directive="10-517",
        channels=[
            "LSR.ALL",
            "LSR.&lt;typetext&gt;",
            "LSR.&lt;state&gt;",
            "LSR.&lt;state&gt;.&lt;typetext&gt;",
        ],
    ),
    dict(
        afos="MCD",
        directive="10-517",
        channels=[C3, C7],
        notes=(
            "The WFOs"
            " included are based on the ones highlighted by SPC within "
            "the text and not from a spatial check of their polygon."
        ),
    ),
    dict(
        afos="MPD",
        directive="10-517",
        channels=[C3, C7],
        notes=(
            "The WFOs"
            " included are based on the ones highlighted by WPC within "
            "the text and not from a spatial check of their polygon."
        ),
    ),
    dict(afos="MIS", directive="10-1701", channels=S2),
    dict(afos="MWS", directive="10-314", channels=S2),
    dict(afos="NOW", directive="10-517", channels=S2),
    dict(afos="NSH", directive="10-1701", channels=S2),
    dict(afos="OAV", directive="10-1701", channels=S2),
    dict(afos="OMR", directive="10-1701", channels=S2),
    dict(afos="PFM", directive="10-1701", channels=S2),
    dict(afos="PNS", directive="10-1701", channels=S2),
    dict(afos="PSH", directive="10-1701", channels=S2),
    dict(
        afos="PFWFD1",
        directive="10-512",
        channels=[
            C1,
            "&lt;wfo&gt;.SPCFD1",
            "&lt;wfo&gt;.SPCFD1.&lt;threshold&gt;",
        ],
        notes=("There is no present means for county/zone based channels."),
    ),
    dict(
        afos="PFWFD2",
        directive="10-512",
        channels=[
            C1,
            "&lt;wfo&gt;.SPCFD2",
            "&lt;wfo&gt;.SPCFD2.&lt;threshold&gt;",
        ],
        notes=("There is no present means for county/zone based channels."),
    ),
    dict(
        afos="PTSDY1",
        directive="10-512",
        channels=[
            C1,
            "&lt;wfo&gt;.SPCDY1",
            "&lt;wfo&gt;.SPCDY1.&lt;threshold&gt;",
        ],
        notes=("There is no present means for county/zone based channels."),
    ),
    dict(
        afos="PTSDY2",
        directive="10-512",
        channels=[
            C1,
            "&lt;wfo&gt;.SPCDY2",
            "&lt;wfo&gt;.SPCDY2.&lt;threshold&gt;",
        ],
        notes=("There is no present means for county/zone based channels."),
    ),
    dict(afos="REC", directive="10-1701", channels=S2),
    dict(afos="RER", directive="10-1004", channels=S2),
    dict(
        afos="RBG94E",
        directive="10-930",
        channels=[
            C1,
            "&lt;wfo&gt;.ERODY1",
            "&lt;wfo&gt;.ERODY1.&lt;threshold&gt;",
        ],
        notes=("There is no present means for county/zone based channels."),
    ),
    dict(
        afos="RBG98E",
        directive="10-930",
        channels=[
            C1,
            "&lt;wfo&gt;.ERODY2",
            "&lt;wfo&gt;.ERODY2.&lt;threshold&gt;",
        ],
        notes=("There is no present means for county/zone based channels."),
    ),
    dict(
        afos="RBG99E",
        directive="10-930",
        channels=[
            C1,
            "&lt;wfo&gt;.ERODY3",
            "&lt;wfo&gt;.ERODY3.&lt;threshold&gt;",
        ],
        notes=("There is no present means for county/zone based channels."),
    ),
    dict(afos="RRM", directive="10-1701", channels=S2),
    dict(afos="RFD", directive="10-1701", channels=S2),
    dict(afos="RTP", directive="10-1701", channels=S2),
    dict(afos="RVA", directive="10-1701", channels=S2),
    dict(afos="RVD", directive="10-922", channels=S2),
    dict(afos="RVF", directive="10-912", channels=S2),
    dict(afos="RWS", directive="10-1701", channels=S2),
    dict(afos="RVS", directive="10-1701", channels=S2),
    dict(afos="SAB", directive="10-1701", channels=S2),
    dict(afos="STF", directive="10-1701", channels=S2),
    dict(afos="SPS", directive="10-517", channels=[C3, C3p, C3S, C6, C9]),
    dict(afos="SRF", directive="10-1701", channels=S2),
    dict(afos="SPW", directive="10-1701", channels=S2),
    dict(afos="TAF", directive="10-1701", channels=[C3, C3p, C8]),
    dict(afos="TCD", directive="10-1701", channels=S21),
    dict(afos="TCM", directive="10-1701", channels=S21),
    dict(afos="TCP", directive="10-1701", channels=S21),
    dict(afos="TCU", directive="10-1701", channels=S21),
    dict(afos="TIB", directive="10-1701", channels=S2),
    dict(afos="TID", directive="10-320", channels=S2),
    dict(afos="TOE", directive="10-1701", channels=S2),
    dict(afos="TWO", directive="10-1701", channels=S21),
    dict(afos="WSV", directive="10-1701", channels=S2),
    dict(afos="VAA", directive="10-1701", channels=S2),
    dict(afos="WRK", directive="10-1701", channels=S2),
    dict(afos="ZFP", directive="10-1701", channels=S2),
]


def get_data(afos):
    """Return the text data for a given afos identifier"""
    fn = f"../examples/{afos}.txt"
    if not os.path.isfile(fn):
        print(f"File {fn} is missing")
    with open(fn, "rb") as fh:
        data = fh.read()
    return (
        data.decode("ascii")
        .replace("\r", "")
        .replace("\001\n", "")
        .replace("\003", "")
    )


def load_dicts():
    """Load up the directionaries"""
    pgconn, cursor = get_dbconnc("postgis")
    sql = """
        SELECT name, ugc, wfo from ugcs WHERE
        name IS NOT Null and end_ts is null
    """
    cursor.execute(sql)
    for row in cursor:
        nm = (row["name"]).replace("\x92", " ").replace("\xc2", " ")
        wfos = re.findall(r"([A-Z][A-Z][A-Z])", row["wfo"])
        ugc_dict[row["ugc"]] = UGC(
            row["ugc"][:2], row["ugc"][2], row["ugc"][3:], name=nm, wfos=wfos
        )

    sql = """SELECT nwsli,
     river_name || ' ' || proximity || ' ' || name || ' ['||state||']' as rname
     from hvtec_nwsli"""
    cursor.execute(sql)
    for row in cursor:
        nm = row["rname"].replace("&", " and ")
        nwsli_dict[row["nwsli"]] = NWSLI(row["nwsli"], name=nm)

    nwsli_dict["MLU"] = NWSLI("MLU", lat=32.52, lon=-92.03)
    nwsli_dict["IGB"] = NWSLI("MLU", lat=33.48, lon=-88.52)
    nwsli_dict["MEI"] = NWSLI("MLU", lat=32.38, lon=-88.80)


def do_generic(fh):
    """Handle the generic case."""
    fh.write(
        """
    <h3>NWS Local Office / National Products</h3>
    <div class="table-responsive">
    <table class="table table-bordered table-sm">
    <thead class="table-light">
    <tr>
      <th scope="col" class="text-center"> </th>
      <th scope="col">AFOS PIL + Product Name</th>
      <th scope="col">Directive</th>
      <th scope="col">Channel Templates Used</th>
    </tr>
    </thead>
    """
    )
    for entry in GEN_PRODUCTS:
        afos: str = entry["afos"]
        if afos == "":
            continue
        try:
            v = PARSERS.get(afos, productparser)(
                get_data(afos),
                ugc_provider=ugc_dict,
                nwsli_provider=nwsli_dict,
            )
            assert v.afos is not None
        except Exception as exp:
            LOG.info("ABORT: productparser %s failed", afos)
            LOG.exception(exp)
            sys.exit()
        j = v.get_jabbers("https://mesonet.agron.iastate.edu/p.php")
        channels = []
        _, html, xtra = j[0]
        tweet = xtra["twitter"] + "<br />"
        jmsg = html
        if isinstance(xtra["channels"], list):
            xtra["channels"] = ",".join(xtra["channels"])
        for channel in xtra["channels"].split(","):
            if channel not in channels:
                channels.append(channel)
        channels.sort()
        fh.write(
            """<tr>
  <td class="text-center align-middle">
    <button class="btn btn-sm" type="button"
      data-bs-toggle="collapse" data-bs-target="#channel_%s"
      aria-expanded="false">
      <i class="bi bi-plus-lg" aria-hidden="true"></i>
      <span class="visually-hidden">Toggle details</span>
    </button>
  </td>
  <td>%s (%s)</td>
  <td><a href="%s">%s</a></td>
  <td>%s</td>
</tr>
<tr>
  <td colspan="4">
    <div id="channel_%s" class="collapse">
    <dl class="row mb-0">
        %s
        <dt>Example Raw Text:</dt>
<dd><a href="https://mesonet.agron.iastate.edu/p.php?pid=%s">View Text</a></dd>
        <dt>Channels for Product Example:</dt><dd>%s</dd>
        <dt>XMPP Chatroom Example:</dt><dd>%s</dd>
        <dt>Twitter Example:</dt><dd>%s</dd>
        </dl>
        </div>
        </td>
        </tr>
        """
            % (
                afos,
                SPECIAL.get(afos, prodDefinitions.get(afos, afos)),
                afos,
                D[entry["directive"]],
                entry["directive"],
                " ".join(
                    [
                        f'<span class="badge bg-light text-dark">{s}</span>'
                        for s in entry["channels"]
                    ]
                ),
                afos,
                "<dt>Notes</dt><dd>%s</dd>" % (entry.get("notes"),)
                if "notes" in entry
                else "",
                v.get_product_id(),
                " ".join(
                    [
                        f'<span class="badge bg-light text-dark">{s}</span>'
                        for s in channels
                    ]
                ),
                jmsg,
                tweet,
            )
        )

    fh.write("""</table></div>""")


def do_vtec(fh):
    """Do VTEC"""
    fh.write(
        """
    <h3>NWS Products with P-VTEC and/or H-VTEC Included</h3>
    <div class="table-responsive">
    <table class="table table-bordered table-sm">
    <thead class="table-light">
    <tr>
      <th scope="col" class="text-center"> </th>
      <th scope="col">AFOS PIL + Product Name</th>
      <th scope="col">Directive</th>
      <th scope="col">Channel Templates Used</th>
    </tr>
    </thead>
    """
    )
    for entry in VTEC_PRODUCTS:
        afos = entry["afos"]
        v = vtec_parser(
            get_data(afos), ugc_provider=ugc_dict, nwsli_provider=nwsli_dict
        )
        if v.afos is None:
            LOG.info("afos was None for %s", entry)
            continue
        j = v.get_jabbers(
            "https://mesonet.agron.iastate.edu/vtec/",
            "https://mesonet.agron.iastate.edu/vtec/",
        )
        jmsg = ""
        tweet = ""
        channels = []
        for _, html, xtra in j:
            tweet += xtra["twitter"] + "<br />"
            jmsg += html
            for channel in xtra["channels"].split(","):
                if channel not in channels:
                    channels.append(channel)
        channels.sort()
        fh.write(
            """<tr>
  <td class="text-center align-middle">
    <button class="btn btn-sm" type="button"
      data-bs-toggle="collapse" data-bs-target="#channel_%s"
      aria-expanded="false">
      <i class="bi bi-plus-lg" aria-hidden="true"></i>
      <span class="visually-hidden">Toggle details</span>
    </button>
  </td>
  <td>%s (%s)</td>
  <td><a href="%s">%s</a></td>
  <td>%s</td>
</tr>
<tr>
  <td colspan="4">
    <div id="channel_%s" class="collapse">
    <dl class="row mb-0">
        %s
        <dt>Example Raw Text:</dt>
<dd><a href="https://mesonet.agron.iastate.edu/p.php?pid=%s">View Text</a></dd>
        <dt>Channels for Product Example:</dt><dd>%s</dd>
        <dt>XMPP Chatroom Example:</dt><dd>%s</dd>
        <dt>Twitter Example:</dt><dd>%s</dd>
        </dl>
        </div>
        </td>
        </tr>
        """
            % (
                afos,
                prodDefinitions.get(afos, afos),
                afos,
                D[entry["directive"]],
                entry["directive"],
                " ".join(
                    [
                        f'<span class="badge bg-light text-dark">{s}</span>'
                        for s in entry["channels"]
                    ]
                ),
                afos,
                "<dt>Notes</dt><dd>%s</dd>" % (entry.get("notes"),)
                if "notes" in entry
                else "",
                v.get_product_id(),
                " ".join(
                    [
                        f'<span class="badge bg-light text-dark">{s}</span>'
                        for s in channels
                    ]
                ),
                jmsg,
                tweet,
            )
        )

    fh.write("""</table></div>""")


def main():
    """Do Something Fun"""
    load_dicts()
    with open(CHANNELSFN, "w", encoding="utf-8") as fh:
        do_vtec(fh)
        do_generic(fh)


if __name__ == "__main__":
    # Go Main Go
    main()



"""FFG"""

# 3rd Party
import click
from pyiem.nws.products.ffg import parser

# Local
from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge


def real_parser(txn, buf):
    """callback func"""
    ffg = parser(buf, ugc_provider={})
    if ffg.afos == "FFGMPD":
        return
    if common.dbwrite_enabled():
        ffg.sql(txn)
    if ffg.warnings and ffg.warnings[0].find("termination") == -1:
        common.email_error("\n".join(ffg.warnings), buf)
    sz = 0 if ffg.data is None else len(ffg.data.index)
    LOG.info("FFG found %s entries for product %s", sz, ffg.get_product_id())


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Our main method"""
    bridge(real_parser, dbpool=get_database("postgis"))



"""Monitor a directory for new files and then do the level II dance"""

import os
import re
import subprocess

import inotify.adapters

FNPATTERN = re.compile("OP5R_[0-9]{8}_[0-9]{4}")


def main():
    """Go Main Go."""
    i = inotify.adapters.Inotify()

    i.add_watch("/home/op5r")
    try:
        for event in i.event_gen():
            if event is not None:
                (_header, type_names, watch_path, filename) = event
                if "IN_CLOSE_WRITE" not in type_names:
                    continue
                fn = filename.decode("utf-8")
                if not FNPATTERN.match(fn):
                    print("invalid filename: %s" % (fn,))
                    continue
                cmd = f"pqinsert -f NEXRAD2 -p '{fn}' {watch_path}/{fn}"
                subprocess.call(cmd, shell=True)
                os.unlink("%s/%s" % (watch_path, fn))

    finally:
        i.remove_watch("/local/op5r")


if __name__ == "__main__":
    main()



"""Test pywwa.common"""

from pyiem.util import utc

from pywwa import CTX, common


def test_setup_syslog():
    """API exercice."""
    CTX["stdout_logging"] = True
    common.setup_syslog()


def test_parse_utcnow():
    """Test that we can parse utcnow."""
    assert common.parse_utcnow("2017-01-01T12:00").year == 2017
    assert common.parse_utcnow(None) is None


def test_crawl():
    """Test crawling before walking."""
    assert common


def test_should_email():
    """Test that our logic to prevent email bombs works."""
    common.SETTINGS["pywwa_email_limit"] = 10
    for _ in range(30):
        common.EMAIL_TIMESTAMPS.append(utc())
    assert not common.should_email()
    assert not common.email_error(None, None)


def test_email_error():
    """Test that we can email an error."""
    common.EMAIL_TIMESTAMPS = []
    common.email_error(None, None)
    CTX["disable_email"] = True
    common.email_error(None, None)



"""Convert noaaport GINI imagery into GIS PNGs

I convert raw GINI noaaport imagery into geo-referenced PNG files both in the
'native' projection and 4326.
"""

# Local
from pywwa.workflows.gini2gis import main

if __name__ == "__main__":
    main()



"""CWA Product Parser!"""

# 3rd Party
import click
from pyiem.nws.products.cwa import parser

# Local
from pywwa import LOG, common, get_data_filepath
from pywwa.database import get_database, get_dbconnc
from pywwa.ldm import bridge

# Load LOCS table
LOCS = {}


def goodline(line):
    """Is this line any good?"""
    return len(line) > 69 and line[0] not in ["!", "#"]


def load_database(txn):
    """Load database stations."""

    txn.execute(
        "SELECT id, name, ST_x(geom) as lon, ST_y(geom) as lat from stations "
        "WHERE network ~* 'ASOS'"
    )
    for row in txn.fetchall():
        LOCS[row["id"]] = row

    fn = get_data_filepath("faa_apt.tbl")
    with open(fn, encoding="utf-8") as fh:
        for line in filter(goodline, fh):
            sid = line[:4].strip()
            lat = float(line[56:60]) / 100.0
            lon = float(line[61:67]) / 100.0
            name = line[16:47].strip()
            if sid not in LOCS:
                LOCS[sid] = {"lat": lat, "lon": lon, "name": name}

    fn = get_data_filepath("vors.tbl")
    with open(fn, encoding="utf-8") as fh:
        for line in filter(goodline, fh):
            sid = line[:3]
            lat = float(line[56:60]) / 100.0
            lon = float(line[61:67]) / 100.0
            name = line[16:47].strip()
            LOCS[sid] = {"lat": lat, "lon": lon, "name": name}

    fn = get_data_filepath("pirep_navaids.tbl")
    with open(fn, encoding="utf-8") as fh:
        for line in filter(goodline, fh):
            sid = line[:3]
            lat = float(line[56:60]) / 100.0
            lon = float(line[61:67]) / 100.0
            LOCS[sid] = {"lat": lat, "lon": lon}
    LOG.info("Loaded %s locations into LOCS", len(LOCS))


def process_data(txn, data):
    """Process the product"""
    prod = parser(
        data, nwsli_provider=LOCS, utcnow=common.utcnow(), ugc_provider={}
    )
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), data)
    prod.sql(txn)
    for j in prod.get_jabbers(
        common.SETTINGS.get("pywwa_product_url", "pywwa_product_url"), ""
    ):
        common.send_message(j[0], j[1], j[2])
    return prod


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Fire things up."""
    conn, cursor = get_dbconnc("mesosite")
    load_database(cursor)
    conn.close()
    bridge(process_data, dbpool=get_database("postgis"))



"""Test fake_afos_dump."""

# 3rd Party
import pytest
from pyiem.nws.products.pirep import Pirep
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import fake_afos_dump


@pytest.mark.parametrize("database", ["afos"])
def test_DSM(cursor):
    """Test fake DSM."""
    tp = fake_afos_dump.really_process_data(
        cursor, get_example_file("DSM.txt")
    )
    # pyiem now sets the afos value when it is none to PIREP
    assert tp.afos == "DSMZME"


def test_fake_pirep():
    """Test fake PIREP."""
    tp = Pirep(get_example_file("PIREP.txt"), utcnow=utc(2024, 7, 9, 0, 0))
    # pyiem now sets the afos value when it is none to PIREP
    assert tp.afos == "PIREP"


@pytest.mark.parametrize("database", ["afos"])
def test_processor(cursor):
    """Test basic parsing."""
    data = get_example_file("CWA.txt")
    pywwa.CTX["utcnow"] = utc(2011, 3, 3, 6, 16)
    tp = fake_afos_dump.really_process_data(cursor, data)
    assert tp.valid == utc(2011, 3, 3, 1, 5)


@pytest.mark.parametrize("database", ["afos"])
def test_fake_noxx(cursor):
    """WMO header faked."""
    data = get_example_file("NOXX.txt")
    pywwa.CTX["utcnow"] = utc(2024, 1, 2, 18, 51)
    tp = fake_afos_dump.really_process_data(cursor, data)
    assert tp.afos == "ADMWNO"



"""
Script that merges the upstream HVTEC NWSLI information into the database. We
are called like so

python merge_hvtec_nwsli.py hvtec_list_10132020.csv

Whereby the argument to the script is the filename stored here:

https://www.weather.gov/vtec/Valid-Time-Event-Code

hvtec_nwsli table:
 nwsli      | character(5)           |
 river_name | character varying(128) |
 proximity  | character varying(16)  |
 name       | character varying(128) |
 state      | character(2)           |
 geom       | geometry               |
"""

# stdlib
import sys

# 3rd Party
import httpx
from pyiem.util import logger

from pywwa.database import get_dbconnc

LOG = logger()


def main(argv) -> int:
    """Go Main"""
    if len(argv) < 2:
        print("USAGE: python merge_hvtec_nwsli.py FILENAME")
        return 1

    dbconn, cursor = get_dbconnc("postgis")
    LOG.warning(" - Connected to database: postgis")

    fn = argv[1]
    uri = f"https://www.weather.gov/media/vtec/{fn}"

    LOG.warning(" - Fetching file: %s", uri)
    req = httpx.get(uri, timeout=30)
    if req.status_code != 200:
        LOG.warning("Got status_code %s for %s", req.status_code, uri)
        return 1
    updated = 0
    new = 0
    bad = 0
    for linenum, line in enumerate(req.content.decode("ascii").split("\n")):
        if line.strip() == "":
            continue
        tokens = line.strip().split(",")
        if len(tokens) != 7:
            LOG.warning(
                " + Linenum %s had %s tokens, instead of 7\n%s",
                linenum + 1,
                len(tokens),
                line,
            )
            bad += 1
            continue
        (nwsli, river_name, proximity, name, state, lat, lon) = tokens
        if "\\N" in [lat, lon] or "" in [lat, lon]:
            LOG.warning(
                " + Linenum %s [%s] had a null lat/lon\n%s",
                linenum + 1,
                nwsli,
                line,
            )
            bad += 1
            continue
        if len(nwsli) != 5:
            LOG.warning(
                ' + Linenum %s had a NWSLI "%s" not of 5 character length\n%s',
                linenum + 1,
                nwsli,
                line,
            )
            bad += 1
            continue
        cursor.execute("DELETE from hvtec_nwsli WHERE nwsli = %s", (nwsli,))
        if cursor.rowcount == 1:
            updated += 1
        else:
            new += 1
        giswkt = f"SRID=4326;POINT({0 - float(lon)} {float(lat)})"
        sql = """
            INSERT into hvtec_nwsli (nwsli, river_name, proximity, name,
             state, geom) values (%s, %s, %s, %s, %s, %s)
             """
        args = (
            nwsli,
            river_name,
            proximity[:16],
            name,
            state,
            giswkt,
        )
        cursor.execute(sql, args)

    cursor.close()
    dbconn.commit()
    LOG.warning(
        " - DONE! %s updated %s new, %s bad entries", updated, new, bad
    )
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""Test dsm_parser."""

# 3rd Party
import pytest
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import dsm


@pytest.mark.parametrize("database", ["mesosite"])
def test_load_stations(cursor):
    """Test station loading."""
    # Need to set one station to a bad tzname to exercise an exception
    cursor.execute(
        "UPDATE stations SET tzname = 'BoG0S' where id = 'MSP' and "
        "network = 'MN_ASOS'"
    )
    assert cursor.rowcount == 1
    dsm.load_stations(cursor)


@pytest.mark.parametrize("database", ["iem"])
def test_processor(cursor):
    """Test basic parsing."""
    data = get_example_file("DSM.txt")
    pywwa.CTX["utcnow"] = utc(2011, 11, 27, 6, 16)
    dsm.real_parser(cursor, data)



"""Poll the OneWire bus for devices."""

import os
import subprocess
import time
from datetime import datetime

os.environ["TZ"] = "CST6CDT"


def work():
    """Do the work."""
    with subprocess.Popen(
        ["./digitemp", "-q", "-a", "-s", "/dev/ttyS0", "-o", "%s %.2F"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as proc:
        stdout, stderr = proc.communicate()
    if stderr != b"":
        print("Rut roh")
        print(stderr.decode("ascii"))
        return
    d = stdout.decode("ascii").split("\n")
    with open("onewire.txt", "w") as fh:
        fh.write("\n".join(d))
    data = {}
    for line in d:
        if not line:
            continue
        t = line.split()
        try:
            data[int(t[0])] = t[1]
        except Exception:
            print(line)
    if len(data) < 4:
        print("ERROR: not enough data")
        return
    now = datetime.now()
    fp = f"ot0003_{now:%Y%m%d%H%M}.dat"
    with open(fp, "w") as fh:
        fh.write(
            f"104,{now:%Y},{now:%j},{now:%H%M},{data[0]}, "
            f"{data[1]}, {data[2]}, {data[3]},11.34\n"
        )
    subprocess.call(["/home/meteor_ldm/bin/pqinsert", fp])
    os.remove(fp)


def main():
    """Go Main Go."""
    with open("runner.pid", "w") as fh:
        fh.write(f"{os.getpid()}")
    while 1:
        work()
        time.sleep(58)


if __name__ == "__main__":
    main()



"""Test Database Functionality."""

# Third Party
from psycopg.rows import dict_row

# Local
from pywwa import database


def test_database():
    """Test that we can call the API."""
    database.load_nwsli({})


def test_load_metar_stations():
    """Test loading of METAR stations."""
    pgconn = database.get_dbconn("mesosite")
    cursor = pgconn.cursor(row_factory=dict_row)
    database.load_metar_stations(cursor, {})
    pgconn.close()



"""Fake a NOAAPort product for WPC ERO.

This is no longer used for ingest of geometries, but the text file is still
processed for archival.
"""

import os
import subprocess
import tempfile

import httpx
from pyiem.database import get_dbconn
from pyiem.nws.product import TextProduct
from pyiem.util import (
    exponential_backoff,
    logger,
    noaaport_text,
    utc,
)

LOG = logger()
BASEURL = "https://www.wpc.ncep.noaa.gov/qpf/"
QUEUE = {
    "94epoints.txt": ["MENC98", "RBG94E"],
    "94epoints_remaining.txt": ["MENC98", "RBG94E"],
    "98epoints.txt": ["MENU98", "RBG98E"],
    "99epoints.txt": ["MENU98", "RBG99E"],
}


def run(cursor, fn, ttaaii, awipsid):
    """Make things happen."""
    cursor.execute(
        "select entered from products where pil = %s and "
        "entered > now() - '2 days'::interval and entered < now() "
        "ORDER by entered DESC",
        (awipsid,),
    )
    current = utc(1980)
    if cursor.rowcount > 0:
        current = cursor.fetchone()[0]
    req = exponential_backoff(httpx.get, f"{BASEURL}/{fn}", timeout=30)
    if req is None or req.status_code != 200:
        LOG.info("failed to fetch %s", fn)
        return
    data = "\n".join(["000 ", f"{ttaaii} KWNH 010000", awipsid, ""]) + req.text
    # Sometimes we get a NUL character somehow, so we remove it!
    data = data.replace("\000", "")
    tp = TextProduct(data, parse_segments=False, ugc_provider={})
    LOG.debug("For %s current: %s, tp.valid: %s", awipsid, current, tp.valid)
    if tp.valid <= current:
        return
    ts = tp.valid.strftime("%d%H%M")
    data = data.replace("KWNH 010000\n", f"KWNH {ts}\n")
    data = noaaport_text(data)

    tmpfd = tempfile.NamedTemporaryFile(delete=False)
    with open(tmpfd.name, "wb") as fh:
        fh.write(data.encode("ascii"))
    cmd = f"pqinsert -f IDS -p '{ttaaii} KWNH {ts} /p{awipsid}' {tmpfd.name}"
    LOG.debug(cmd)
    subprocess.call(cmd, shell=True)
    os.unlink(tmpfd.name)


def main():
    """Go Main Go."""
    with get_dbconn("afos") as pgconn:
        cursor = pgconn.cursor()
        for fn, (ttaaii, awipsid) in QUEUE.items():
            run(cursor, fn, ttaaii, awipsid)


if __name__ == "__main__":
    main()



"""Test taf_parser."""

# 3rd Party
import pytest

from pywwa.testing import get_example_file

# Local
from pywwa.workflows import taf


@pytest.mark.parametrize("database", ["asos"])
def test_real_parser(cursor):
    """Test that we can parse TAFS!"""
    taf.real_process(cursor, get_example_file("TAF.txt"))



"""Test the CLI main function on all workflows."""

import importlib
import pkgutil

import pytest
from click.testing import CliRunner

import pywwa.workflows
from pywwa.workflows.cli import main as cli_main


def find_mains():
    """yield all main functions within pywwa.workflows."""
    package = pywwa.workflows
    for _importer, modname, _ispkg in pkgutil.iter_modules(package.__path__):
        module = importlib.import_module(package.__name__ + "." + modname)
        for name in dir(module):
            if name.startswith("_"):
                continue
            obj = getattr(module, name)
            if not callable(obj):
                continue
            if not hasattr(obj, "main"):
                continue
            yield obj


@pytest.mark.parametrize("mainmethod", find_mains())
def test_exercise_api(mainmethod):
    """Exercise the main() method via click."""
    runner = CliRunner()
    result = runner.invoke(mainmethod, ["-x"])
    if result.exception:
        raise result.exception
    assert result.exit_code == 0


def test_cli_with_jabber():
    """Test the CLI main function."""
    runner = CliRunner()
    result = runner.invoke(cli_main)
    if result.exception:
        raise result.exception
    assert result.exit_code == 0



"""Feed archived NCR data to nexrad3_attr.py , should be run like

python util/feed_archived_ncr.py NEXRAD YYYY MM
  | YYYY=2009 MM=01 python nexrad3_attr.py
"""

import datetime
import glob
import os
import subprocess
import sys
import time


def main(argv):
    """Go Main Go."""
    if not os.path.isdir("/tmp/l3tmp"):
        os.makedirs("/tmp/l3tmp")

    nexrad = sys.argv[1]

    sts = datetime.datetime(int(argv[2]), int(argv[3]), 1)
    ets = sts + datetime.timedelta(days=32)
    ets = ets.replace(day=1)
    interval = datetime.timedelta(days=1)

    now = sts
    while now < ets:
        afn = now.strftime(
            f"/mesonet/ARCHIVE/nexrad/%Y_%m/{nexrad}_%Y%m%d.tgz"
        )
        if not os.path.isfile(afn):
            sys.stderr.write(f"Missing {afn}\n")
            now += interval
            continue

        subprocess.call(f"tar -x -z -C /tmp/l3tmp -f {afn} NCR", shell=True)
        if not os.path.isdir("/tmp/l3tmp/NCR"):
            sys.stderr.write(f"Missing NCR data for {afn}\n")
            now += interval
            continue

        files = glob.glob("/tmp/l3tmp/NCR/NCR_*")
        for fn in files:
            # Need fake seq id
            with open(fn, "rb") as fh:
                data = fh.read()
            if len(data) < 20:
                sys.stderr.write(f"short read on {fn}\n")
                continue
            if data[0] == "S":
                sys.stdout.write("\001\r\r\n000\r\r\n")
            sys.stdout.write(data)
            if data[-4:] != "\r\r\n\003":
                sys.stdout.write("\r\r\n\003")
            time.sleep(0.25)

        if os.path.isdir("/tmp/l3tmp/NCR"):
            subprocess.call("rm -rf /tmp/l3tmp/NCR", shell=True)

        now += interval


if __name__ == "__main__":
    main(sys.argv)



"""Test spe_parser."""

# 3rd Party
import pytest
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import spe


@pytest.mark.parametrize("database", ["postgis"])
def test_processor(cursor):
    """Test basic parsing."""
    data = get_example_file("SPE.txt")
    pywwa.CTX["utcnow"] = utc(2011, 10, 31, 1, 16)
    spe.real_process(cursor, data)



"""test import failure via a mock."""

import sys

import pywwa


class ImportErrorWhenAccessed:
    def __getattr__(self, name):
        raise ImportError


def test_import_error():
    """Test that we can handle an import error."""
    pywwa.SETTINGS["__setup_syslog"] = False
    from pywwa.common import setup_syslog

    for name in ("twisted.python", "twisted.python.syslog"):
        del sys.modules[name]

    sys.modules["twisted.python"] = ImportErrorWhenAccessed()
    setup_syslog()
    del sys.modules["twisted.python"]



"""Support lib for the parser scripts found in this directory"""

import datetime
import getpass
import inspect
import logging
import os
import socket
import sys
import traceback
from email.mime.text import MIMEText
from io import StringIO

import click
import pyiem
from pybufrkit.bufr import log as bufrlog
from pyiem.util import utc
from twisted.internet import reactor
from twisted.mail import smtp
from twisted.python import failure
from twisted.python import log as tplog

# Local Be careful of circeref here
from pywwa import CTX, LOG, SETTINGS
from pywwa.database import get_dbconn

# http://bugs.python.org/issue7980
datetime.datetime.strptime("2013", "%Y")
EMAIL_TIMESTAMPS = []


class CustomFormatter(logging.Formatter):
    """A custom log formatter class."""

    def format(self, record):
        """Return a string!"""
        return (
            f"[{record.filename}:{record.lineno} {record.funcName}] "
            f"{record.getMessage()}"
        )


def utcnow():
    """Return what utcnow is based on command line."""
    return utc() if CTX["utcnow"] is None else CTX["utcnow"]


def dbwrite_enabled():
    """Is database writing not-disabled as per command line."""
    return not CTX["disable_dbwrite"]


def replace_enabled():
    """Is -r --replace enabled."""
    return CTX["replace"]


def setup_syslog():
    """Setup how we want syslogging to work"""
    # I am punting on some issue the below creates within pytest
    if SETTINGS.get("__setup_syslog", False):
        return

    loglvl = logging.DEBUG if sys.stdout.isatty() else logging.INFO

    # https://stackoverflow.com/questions/13699283
    frame = inspect.stack()[-1]
    module = inspect.getmodule(frame[0])
    filename = (
        "None" if module is None else os.path.basename(str(module.__file__))
    )
    # Set the system name for Twisted logs
    system_name = filename.replace("pywwa-parse-", "")

    # Log stuff to stdout if we are running from command line.
    if CTX["stdout_logging"]:
        tplog.addObserver(tplog.FileLogObserver(sys.stdout).emit)

    # windows does not have syslog
    try:
        from twisted.python import syslog

        syslog.startLogging(
            prefix="pyWWA",
            facility=syslog.syslog.LOG_LOCAL2,
            setStdout=not CTX["stdout_logging"],
        )
    except ImportError as exp:
        tplog.msg("Failed to import twisted.python.syslog", system=system_name)
        tplog.err(exp, system=system_name)

    # Add a handler to forward pyIEM logs to Twisted's logging system
    class TwistedLogHandler(logging.Handler):
        """Custom handler to forward logs to Twisted's logging system."""

        def emit(self, record):
            # Use the logging formatter to format the record
            msg = self.format(record)
            tplog.msg(
                msg, logLevel=record.levelname.lower(), system=system_name
            )

    # Create the handler and set the formatter
    twisted_handler = TwistedLogHandler()
    twisted_handler.setFormatter(CustomFormatter())

    # Prevent some noisy INFO messages from pybufrkit
    bufrlog.setLevel(logging.WARNING)

    root_logger = logging.getLogger()  # Get the root logger
    root_logger.setLevel(loglvl)
    root_logger.addHandler(twisted_handler)

    # Prevent a double setup of syslog
    SETTINGS["__setup_syslog"] = True


def load_settings():
    """Load database properties."""
    with get_dbconn("mesosite") as dbconn:
        cursor = dbconn.cursor()
        cursor.execute("SELECT propname, propvalue from properties")
        for row in cursor:
            SETTINGS[row[0]] = row[1]
        LOG.info("Loaded %s settings from database", len(SETTINGS))
        cursor.close()


def should_email():
    """Prevent email bombs

    Use the setting `pywwa_email_limit` to threshold the number of emails
    permitted within the past hour

    @return boolean if we should email or not
    """
    EMAIL_TIMESTAMPS.insert(0, utc())
    delta = EMAIL_TIMESTAMPS[0] - EMAIL_TIMESTAMPS[-1]
    email_limit = int(SETTINGS.get("pywwa_email_limit", 10))
    if len(EMAIL_TIMESTAMPS) < email_limit:
        return True
    while len(EMAIL_TIMESTAMPS) > email_limit:
        EMAIL_TIMESTAMPS.pop()

    return delta > datetime.timedelta(hours=1)


def email_error(exp, message="", trimstr=100):
    """
    Helper function to generate error emails when necessary and hopefully
    not flood!
    @param exp A string or perhaps a twisted python failure object
    @param message A string of more information to pass along in the email
    @return boolean If an email was sent or not...
    """
    # Always log a message about our fun
    cstr = StringIO()
    if isinstance(exp, failure.Failure):
        exp.printTraceback(file=cstr)
        LOG.error(exp)
    elif isinstance(exp, Exception):
        traceback.print_exc(file=cstr)
        LOG.error(exp)
    else:
        LOG.info(exp)
    cstr.seek(0)
    if isinstance(message, str):
        LOG.info(message[:trimstr])
    else:
        # Message is bytes and likely not all that useful, so repr and trim it
        message = repr(message)[:trimstr]
        LOG.info(message)

    # Logic to prevent email bombs
    if not should_email():
        LOG.info(
            "Email threshold of %s exceeded, so no email sent!",
            SETTINGS.get("pywwa_email_limit", 10),
        )
        return False

    hn = socket.gethostname()
    hh = f"{getpass.getuser()}@{hn}"
    la = " ".join([f"{a:.2f}" for a in os.getloadavg()])
    txt = (
        f"System          : {hh} [CWD: {os.getcwd()}]\n"
        f"pyiem.version   : {pyiem.__version__}\n"
        f"System UTC date : {utc()}\n"
        f"pyWWA UTC date  : {utcnow()}\n"
        f"process id      : {os.getpid()}\n"
        f"system load     : {la}\n"
        f"Exception       : {exp}\n"
        f"Message:\n{message}\n"
    )
    # prevent any noaaport text from making ugly emails
    msg = MIMEText(txt.replace("\r\r\n", "\n"), "plain", "utf-8")
    # Send the email already!
    msg["subject"] = (
        f"[pyWWA] {sys.argv[0].rsplit('/', maxsplit=1)[-1]} Traceback -- {hn}"
    )
    msg["From"] = SETTINGS.get("pywwa_errors_from", "ldm@localhost")
    msg["To"] = SETTINGS.get("pywwa_errors_to", "ldm@localhost")
    if not CTX["disable_email"]:
        df = smtp.sendmail(
            SETTINGS.get("pywwa_smtp", "smtp"), msg["From"], msg["To"], msg
        )
        df.addErrback(LOG.error)
    else:
        LOG.info("Sending email disabled by command line `-e` flag.")
    return True


def send_message(plain, text, extra):
    """Helper to connect with running JABBER instance."""
    if CTX["JABBER"] is None:
        LOG.info("failed to send as pywwa.JABBER is None, not setup?")
        return
    CTX["JABBER"].send_message(plain, text, extra)


def parse_utcnow(text) -> datetime.datetime:
    """Parse a string into a datetime object."""
    if text is None:
        return None
    fmt = "%Y-%m-%dT%H:%M" if len(text) == 16 else "%Y-%m-%dT%H:%M:%SZ"
    dt = datetime.datetime.strptime(text, fmt)
    return dt.replace(tzinfo=datetime.timezone.utc)


def disable_xmpp(f):
    """Decorator to disable XMPP."""
    f.disable_xmpp = True
    return f


def init(f):
    """Decorator to setup all things."""

    @click.option(
        "-d",
        "--disable-dbwrite",
        is_flag=True,
        help="Disable writing to the database",
    )
    @click.option(
        "-e",
        "--disable-email",
        is_flag=True,
        help="Disable sending emails",
    )
    @click.option(
        "-l",
        "--stdout-logging",
        is_flag=True,
        help="Log to stdout",
    )
    @click.option(
        "-r",
        "--replace",
        is_flag=True,
        help="Replace existing database entries",
    )
    @click.option(
        "-s",
        "--shutdown-delay",
        type=int,
        default=5,
        help="Shutdown after N seconds",
    )
    @click.option(
        "-u",
        "--utcnow",
        help="Define current UTC timestamp",
        type=parse_utcnow,
    )
    @click.option(
        "-x",
        "--disable-xmpp",
        default=getattr(f, "disable_xmppp", False),
        is_flag=True,
        help="Disable XMPP messaging",
    )
    def decorated_function(*args, **kwargs):
        """Decorated function."""
        # Step 1, parse command line arguments into running context
        CTX.update(kwargs)
        # Step 2, setup logging
        setup_syslog()
        # Step 3, load settings from database, which is blocking
        load_settings()
        # Step 4, setup jabber client
        if not CTX["disable_xmpp"] and not getattr(f, "disable_xmpp", False):
            from pywwa.xmpp import make_jabber_client

            make_jabber_client()
        # Step 5, call the function
        ret = f(*args, **kwargs)
        # Step 6, run the reactor, but account for ugliness of pytest
        if not reactor.running:
            reactor.run()
        return ret

    return decorated_function



"""Test spammer."""

import base64

from pyiem.nws.product import TextProduct

from pywwa.testing import get_example_file

# Local
from pywwa.workflows import spammer


def test_tornado_multi():
    """Test the parsing of EF scale."""
    prod = TextProduct(get_example_file("PNS_damage_multi2.txt"))
    res = spammer.damage_survey_pns(prod)
    pos = 0
    # Ensure order
    msg = base64.b64decode(res[2].get_payload()).decode("utf-8")
    for ef in range(3):
        thispos = msg.find(f"EF-{ef}")
        assert thispos > pos
        pos = thispos


def test_tornado_damage():
    """Test the parsing of EF scale."""
    prod = TextProduct(get_example_file("PNS_damage.txt"))
    res = spammer.damage_survey_pns(prod)
    assert "EF2" in res[0]


def test_spammer():
    """Test that we can delineate PNS!"""
    res = spammer.real_process(get_example_file("PNS_damage.txt"))
    assert res is not None
    res = spammer.real_process(get_example_file("PNS.txt"))
    assert res is None



"""MOS Data Ingestor, why not?"""

# 3rd Party
import click
from pyiem.nws.products.mos import parser

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def process_data(txn, text):
    """The real processor of the raw data, fun!"""
    prod = parser(text, utcnow=common.utcnow(), ugc_provider={})
    if not common.dbwrite_enabled():
        return
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), text)
    prod.sql(txn)


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(process_data, dbpool=get_database("mos", cp_max=5))



"""Test the parsers/watch_parser.py module."""

import pytest

from pywwa.testing import get_example_file
from pywwa.workflows import watch


@pytest.mark.parametrize("database", ["postgis"])
def test_parser(cursor):
    """Send a valid watch product to the parser."""
    for fn in ["SAW.txt", "SEL.txt", "WWP.txt"]:
        prod = watch.real_process(cursor, get_example_file(fn))
        assert prod.warnings == []
        assert watch.QUEUE[169][fn[:3]] is not None
    watch.process_queue()
    assert watch.QUEUE == {}



"""Process WPC Excessive Rainfall Outlook (ERO) GeoJSON.

  - Ingest GeoJSON into database.
  - Submit to LDM for archival.

https://www.wpc.ncep.noaa.gov/exper/eromap/geojson/
"""

import json
import os
import subprocess
import tempfile
import warnings
from copy import deepcopy
from datetime import datetime, timezone

import geopandas as gpd
import httpx
import pandas as pd
import xmpp
from pyiem.database import get_dbconn, get_sqlalchemy_conn
from pyiem.nws.products import ero
from pyiem.util import get_properties, logger, utc

# Stop warnings about pyogrio usage of errors=ignore in to_datetime
# https://github.com/geopandas/pyogrio/issues/344
warnings.simplefilter("ignore", FutureWarning)
LOG = logger()
# Only akami allowed
BASEURI = "https://www.wpc.ncep.noaa.gov/exper/eromap/geojson/"
JABBER = {
    "connection": None,
    "to": None,
}
THRESHOLDS = ["MRGL", "SLGT", "MDT", "HIGH"]


def get_jabberconn():
    """Fetch or initialize a jabber connection."""
    if JABBER["connection"] is not None:
        return JABBER["connection"]
    props = get_properties()
    JABBER["to"] = f"{props['bot.username']}@{props['bot.xmppdomain']}"
    jid = xmpp.protocol.JID(
        f"{props['pywwa_jabber_username']}@{props['pywwa_jabber_domain']}/"
        "ero_ingest"
    )
    JABBER["connection"] = xmpp.Client(
        props.get("pywwa_jabber_host", "localhost"),
        debug=[],
    )
    JABBER["connection"].connect()
    JABBER["connection"].auth(
        user=jid.getNode(),
        password=props.get("pywwa_jabber_password"),
        resource=jid.getResource(),
    )
    return JABBER["connection"]


def compute_cycle(day, valid):
    """Figure out an integer cycle that identifies this product."""
    if day == 1:
        if valid.hour in range(0, 6):
            return 1
        if valid.hour in range(6, 13):
            return 8
        if valid.hour in range(13, 20):
            return 16
    if valid.hour in range(4, 12):
        return 8
    if valid.hour in range(17, 22):
        return 20
    return -1


def fetch_ero(day: int, last_issue: datetime) -> gpd.GeoDataFrame:
    """Get the ERO from the WPC website."""
    # origin is IDP, www is akamai
    gdf = None
    for prefix in ["origin", "www"]:
        url = BASEURI.replace("origin", prefix)
        try:
            gdf = gpd.read_file(
                f"{url}Day{day}_Latest.geojson", engine="pyogrio"
            )
            break
        except Exception as exp:
            LOG.info("Failed to fetch %s: %s", url, exp)
    if gdf is None:
        # Swallow transient errors until we are very stale
        lvl = LOG.info
        if (utc() - last_issue).total_seconds() > 86_400:
            lvl = LOG.warning
        lvl("Failed to fetch Day%s", day)
        return None, None
    # Uppercase all the column names
    gdf.columns = [x.upper() if x != "geometry" else x for x in gdf.columns]
    cols = ["ISSUE_TIME", "START_TIME", "END_TIME"]
    for col in cols:
        gdf[col] = pd.to_datetime(
            gdf[col], format="%Y-%m-%d %H:%M:%S"
        ).dt.tz_localize(timezone.utc)
    # Instead of having an empty geojson, they have a bogus polygon with
    # the properties
    if gdf.empty:
        LOG.info("Empty GeoJSON for day %s", day)
        return None, None
    # Merge geometries
    gdf = gdf.dissolve(
        by=gdf.columns.difference(["geometry"]).to_list(), as_index=False
    )
    # Some value add properties that are useful for downstream processing
    gdf["day"] = day
    gdf["threshold"] = ""
    gdf["level"] = -1
    # Save metdata for first row
    meta = gdf.iloc[0][cols].copy()
    gdf = gdf[gdf["OUTLOOK"] != "None Expected"]
    return gdf, meta


def send_to_ldm(gdf, meta, maxissue, day, cycle):
    """Insert for archival and triggering."""
    if cycle < 0:
        LOG.info("Skipping LDM as this is non-regular cycle (-1)")
        return
    for column in gdf.columns:
        try:
            gdf[column] = gdf[column].dt.strftime("%Y-%m-%dT%H:%M:%SZ")
        except Exception as exp:
            LOG.info("Failed to convert %s: %s", column, exp)

    with tempfile.NamedTemporaryFile(delete=False) as tmpfn:
        if gdf.empty:
            tmpfn.write(b'{"type": "FeatureCollection", "features": []}')
        else:
            gdf.to_file(tmpfn.name, driver="GeoJSON")
    # Hack back in some properties
    with open(tmpfn.name, encoding="utf-8") as tmpfh:
        jdict = json.load(tmpfh)
    jdict["iem_properties"] = meta.to_dict()
    for key, value in jdict["iem_properties"].items():
        try:
            jdict["iem_properties"][key] = value.strftime("%Y-%m-%dT%H:%M:%SZ")
        except Exception:
            LOG.info("Failed to convert %s: %s", key, value)

    # Write back out
    with open(tmpfn.name, "w", encoding="utf-8") as tmpfh:
        json.dump(jdict, tmpfh)

    cmd = [
        "pqinsert",
        "-i",
        "-p",
        (
            f"data ac {maxissue:%Y%m%d%H%M} gis/geojson/wpc_ero/"
            f"eroday{day}.geojson GIS/wpc_ero/eroday{day}_{cycle}z.geojson "
            "geojson"
        ),
        tmpfn.name,
    ]
    LOG.info(" ".join(cmd))
    subprocess.call(cmd)
    os.unlink(tmpfn.name)


def get_threshold(threshold):
    """Convert to key."""
    thres = threshold.lower()
    if thres.startswith("slight"):
        return "SLGT"
    if thres.startswith("moderate"):
        return "MDT"
    if thres.startswith("high"):
        return "HIGH"
    if thres.startswith("marginal"):
        return "MRGL"
    LOG.info("Failed to convert '%s' to threshold", threshold)
    return None


def save_df(cursor, gdf: pd.DataFrame, meta, day, cycle):
    """Save to the database."""
    # Create the identifier for this outlook
    cursor.execute(
        "INSERT into spc_outlook(issue, product_issue, expire, updated, "
        "product_id, outlook_type, day, cycle) VALUES "
        "(%s, %s, %s, now(), %s, 'E', %s, %s) RETURNING id",
        (
            meta["START_TIME"],
            meta["ISSUE_TIME"],
            meta["END_TIME"],
            "GEOJSON",
            day,
            cycle,
        ),
    )
    oid = cursor.fetchone()[0]
    # Compute thresholds and levels
    for idx, row in gdf.iterrows():
        threshold = get_threshold(row["OUTLOOK"])
        if threshold is None:
            continue
        gdf.at[idx, "level"] = THRESHOLDS.index(threshold)
        gdf.at[idx, "threshold"] = threshold

    for level, threshold in enumerate(THRESHOLDS):
        larger = gdf[gdf["level"] == level]
        if larger.empty:
            LOG.info("No rows found for %s[%s]", level, THRESHOLDS[level])
            continue
        row = larger.iloc[0]
        smaller = gdf[gdf["level"] == (level + 1)]
        poly = deepcopy(row["geometry"])
        if not smaller.empty:
            smaller = smaller.iloc[0]["geometry"]
            poly = row["geometry"].union(smaller.buffer(0.01))
            LOG.info(
                "Differenced areas poly: %.4f larger: %.4f smaller: %.4f",
                poly.area,
                row["geometry"].area,
                smaller.area,
            )

        LOG.info(
            "Adding %s[%s] %s lrysize: %.4f size: %.4f",
            day,
            cycle,
            threshold,
            row["geometry"].area,
            poly.area,
        )
        cursor.execute(
            """
            INSERT into spc_outlook_geometries (spc_outlook_id, threshold,
            category, geom, geom_layers) VALUES (%s, %s, 'CATEGORICAL',
            ST_SetSrid(ST_Multi(ST_GeomFromEWKT(%s)), 4326),
            ST_SetSrid(ST_Multi(ST_GeomFromEWKT(%s)), 4326))
            """,
            (
                oid,
                threshold,
                row["geometry"].wkt,
                poly.wkt,
            ),
        )


def send_jabber(gdf, issue, day):
    """Do the jabber message generation,"""
    # Create a faked outlook_collections list
    outlook_collections = {
        day: ero.OutlookCollection(
            "issue unused",
            "expire unused",
            day,
        )
    }
    for _, row in gdf.iterrows():
        outlook = ero.Outlook(
            "CATEGORICAL",
            row["threshold"],
            row["geometry"],
        )
        outlook_collections[day].outlooks.append(outlook)

    ero.compute_wfos(outlook_collections)
    jmsgs = ero.jabber_messages(
        issue,
        outlook_collections,
    )
    LOG.info("Sending %d jabber messages", len(jmsgs))
    for txt, html, xtra in jmsgs:
        conn = get_jabberconn()
        msg = xmpp.protocol.Message(
            to=JABBER["to"],
            body=txt,
        )
        xhtml = xmpp.Node(
            "html", {"xmlns": "http://jabber.org/protocol/xhtml-im"}
        )
        xhtml.addChild(
            node=xmpp.simplexml.XML2Node(
                "<body xmlns='http://www.w3.org/1999/xhtml'>"
                + html
                + "</body>"
            )
        )
        msg.addChild(node=xhtml)
        xbot = xmpp.Node(
            "x",
            {
                "xmlns": "nwschat:nwsbot",
                "channels": ",".join(xtra["channels"]),
                "twitter_media": xtra["twitter_media"],
                "twitter": xtra["twitter"],
            },
        )
        # To prevent a bomb against iembot, we need to seed the cache by
        # requesting these twitter_media URLs
        try:
            LOG.info("Fetching %s", xtra["twitter_media"])
            # 30 seconds found to be not quite enough
            httpx.get(xtra["twitter_media"], timeout=45)
        except Exception as exp:
            print(exp)
        msg.addChild(node=xbot)
        conn.send(msg)


def main():
    """Go Main Go."""
    # Load up the most recently processed data.
    with get_sqlalchemy_conn("postgis") as pgconn:
        current = pd.read_sql(
            "SELECT day, "
            "max(product_issue at time zone 'UTC') as last_product_issue from "
            "spc_outlook where outlook_type = 'E' "
            "and product_issue > now() - '7 days'::interval "
            "GROUP by day ORDER by day ASC",
            pgconn,
            index_col="day",
        )
    for day in range(1, 6):
        last_issue = utc(1980)  # default ancient
        if day in current.index:
            # Naive, convert to aware
            last_issue = current.at[day, "last_product_issue"].replace(
                tzinfo=timezone.utc
            )
        # Fetch ERO
        gdf, meta = fetch_ero(day, last_issue)
        if gdf is None:
            continue
        # Compare with what we have, abort if same or older
        maxissue = meta["ISSUE_TIME"]
        if last_issue >= maxissue:
            LOG.info("%s is old %s >= %s", day, last_issue, maxissue)
            continue
        cycle = compute_cycle(day, maxissue)
        # Save to Database
        with get_dbconn("postgis") as dbconn:
            cursor = dbconn.cursor()
            save_df(cursor, gdf, meta, day, cycle)
            cursor.close()
        # Resave to GeoJson without pretty print and decreased precision.
        send_to_ldm(gdf, meta, maxissue, day, cycle)
        # Send Jabber Messages
        send_jabber(gdf, maxissue, day)


if __name__ == "__main__":
    main()



"""SPC Geo Products Parser!"""

# 3rd Party
import click
from pyiem.nws.products.spcpts import parser

# Local
from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge

WAITFOR = 20


def real_parser(txn, buf):
    """Actually process"""
    prod = parser(buf, ugc_provider={})
    if common.dbwrite_enabled():
        prod.sql(txn)
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), buf)
    return prod


def do_jabber(prod):
    """Callback after database work is done."""
    jmsgs = prod.get_jabbers("")
    for txt, html, xtra in jmsgs:
        common.send_message(txt, html, xtra)
    LOG.info(
        "Sent %s messages for product %s", len(jmsgs), prod.get_product_id()
    )


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(real_parser, dbpool=get_database("postgis"), cb2=do_jabber)



"""Test metar_parser."""

# 3rd Party
import pytest
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import metar


def create_entries(cursor):
    """Return a disposable database cursor."""
    # Create fake station, so we can create fake entry in summary
    # and current tables
    cursor.execute(
        "INSERT into stations(id, network, iemid, tzname) "
        "VALUES ('QQQQ', 'FAKE', -1, 'America/Chicago')"
    )
    cursor.execute(
        "INSERT into current(iemid, valid) VALUES (-1, '2015-09-01 00:00+00')"
    )
    cursor.execute(
        "INSERT into summary_2015(iemid, day) VALUES (-1, '2015-09-01')"
    )


def test_api():
    """Test that we can load things to ignore."""
    metar.load_ignorelist()
    metar.cleandb()


@pytest.mark.parametrize("database", ["iem"])
def test_processor(cursor):
    """Test basic parsing."""
    create_entries(cursor)
    data = get_example_file("METAR.txt")
    pywwa.CTX["utcnow"] = utc(2011, 11, 25, 14)
    metadata = {"iemid": -1, "tzname": "America/Chicago"}
    for mtr in metar.process_data(data).metars:
        metar.do_db(cursor, mtr, metadata)



"""Alaska CWF/OFF Parser that generates Fake VTEC."""

import click
from psycopg.connection import Transaction
from pyiem.nws.product import TextProduct

from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge

HEADLINE2VTEC = {
    "BRISK WIND ADVISORY": "BW.Y",
    "GALE WARNING": "GL.W",
    "STORM WARNING": "SR.W",
    "HURRICANE FORCE WIND WARNING": "HF.W",
    "HEAVY FREEZING SPRAY WARNING": "UP.W",
    "SMALL CRAFT ADVISORY": "SC.Y",
}
VTEC_DOMAIN = HEADLINE2VTEC.values()


def get_current(txn: Transaction, nws: TextProduct) -> dict:
    """Get our current entries."""
    wfo = nws.source[1:]
    # VTEC storage does not cross year boundaries, at this time
    table = f"warnings_{nws.valid.year}"
    # Build out the current entries by looking at all UGCs in this product
    current = {}
    ugcdomain = [str(u) for seg in nws.segments for u in seg.ugcs]
    txn.execute(
        f"select ugc, phenomena, significance, eventid from {table}"
        " where wfo = %s and ugc = ANY(%s) and expire > %s and issue <= %s",
        (
            wfo,
            ugcdomain,
            nws.valid,
            nws.valid,
        ),
    )
    for row in txn.fetchall():
        key = f"{row['phenomena']}.{row['significance']}"
        if key not in VTEC_DOMAIN:
            continue
        current[f"{row['ugc']}.{key}"] = row["eventid"]
    return current


def compute_ps(headline: str) -> tuple[str | None, str | None]:
    """Figure out the phenomena and significance from the headline."""
    for key in HEADLINE2VTEC:
        if headline.startswith(key):
            return HEADLINE2VTEC[key].split(".")
    LOG.info("Failed to find PS for %s", headline)
    return None, None


def process_data(txn: Transaction, buf: str) -> TextProduct | None:
    """Actually do something with the buffer, please"""
    if buf.strip() == "":
        return None

    nws = TextProduct(buf, utcnow=common.utcnow())
    wfo = nws.source[1:]
    # VTEC storage does not cross year boundaries, at this time
    table = f"warnings_{nws.valid.year}"

    created_etns = {}
    current = get_current(txn, nws)

    LOG.info(
        "--Processing %s, currently %s entries",
        nws.get_product_id(),
        len(current),
    )
    for segment in nws.segments:
        if len(segment.ugcs) != 1:
            continue
        ugc = segment.ugcs[0]
        ugcexpire = segment.ugcexpire
        for headline in segment.headlines:
            phenomena, significance = compute_ps(headline)
            if phenomena is None:
                continue
            current_key = f"{ugc}.{phenomena}.{significance}"
            # 1. We have already issued, we should extend it
            eventid = current.pop(current_key, None)
            if eventid is not None:
                LOG.info("Extending %s to %s", current_key, ugcexpire)
                txn.execute(
                    f"""
update {table} SET status = 'EXT', expire = %s, updated = %s,
product_ids = array_append(product_ids, %s) WHERE wfo = %s and ugc = %s
and phenomena = %s and significance = %s and eventid = %s
                """,
                    (
                        ugcexpire,
                        nws.valid,
                        nws.get_product_id(),
                        wfo,
                        str(ugc),
                        phenomena,
                        significance,
                        eventid,
                    ),
                )
                continue
            # 2. We have not issued, we should issue
            created_etns_key = f"{phenomena}.{significance}"
            LOG.info("Issuing %s till %s", current_key, ugcexpire)
            if created_etns_key not in created_etns:
                txn.execute(
                    f"""
select coalesce(max(eventid), 0) + 1 as etn from {table}
WHERE wfo = %s and phenomena = %s and significance = %s
                """,
                    (wfo, phenomena, significance),
                )
                row = txn.fetchone()
                etn = row["etn"]
                created_etns[created_etns_key] = etn
            etn = created_etns[created_etns_key]
            txn.execute(
                f"""
INSERT into {table} (issue, expire, updated, wfo, eventid, status,
fcster, ugc, phenomena, significance, gid, init_expire, product_issue,
is_emergency, is_pds, purge_time, product_ids, vtec_year) VALUES
(%s, %s, %s, %s, %s, 'NEW', %s, %s, %s, %s, get_gid(%s, %s),
%s, %s, 'f', 'f', %s, %s, %s)
                """,
                (
                    nws.valid,
                    ugcexpire,
                    nws.valid,
                    wfo,
                    etn,
                    nws.get_signature(),
                    str(ugc),
                    phenomena,
                    significance,
                    str(ugc),
                    nws.valid,
                    ugcexpire,
                    nws.valid,
                    ugcexpire,
                    [nws.get_product_id()],
                    nws.valid.year,
                ),
            )
    for key, eventid in current.items():
        # Cancel
        LOG.info("Canceling [%s]%s", key, eventid)
        (ugc, phenomena, significance) = key.split(".")
        txn.execute(
            f"""
UPDATE {table} SET status = 'CAN', expire = %s, updated = %s,
product_ids = array_append(product_ids, %s)
WHERE wfo = %s and ugc = %s
and phenomena = %s and significance = %s and eventid = %s
""",
            (
                nws.valid,
                nws.valid,
                nws.get_product_id(),
                wfo,
                ugc,
                phenomena,
                significance,
                eventid,
            ),
        )
    return nws


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Fire up our workflow."""
    # Keep things sequential
    bridge(process_data, dbpool=get_database("postgis", cp_max=1))



"""METAR product ingestor

NOTE: It is difficult to keep track of where I am getting the `Metar` library.
So let us document it here for my own sanity.

18 Jul 2017: `snowdepth` branch of my python-metar fork installed with pip
"""

# Local
from pywwa.workflows.metar import main

if __name__ == "__main__":
    main()



"""Sometimes LDM runs dry, but the data is on the website.

crontab job run at :55 after for the previous hour.
"""

import os
import subprocess
from datetime import timedelta

import httpx
from pyiem.util import logger, utc

LOG = logger()


def main():
    """Go Main Go."""
    os.chdir("/mesonet/data/madis/hfmetar")
    now = utc() - timedelta(hours=1)
    testfn = f"{now:%Y%m%d_%H}00_0.nc"
    if os.path.isfile(testfn):
        return
    url = (
        "https://madis-data.cprk.ncep.noaa.gov/madisPublic1/data/LDAD/hfmetar/"
        f"netCDF/{now:%Y%m%d_%H}00.gz"
    )
    LOG.info("fetching %s", url)
    resp = httpx.get(url, timeout=30)
    if resp.status_code != 200:
        LOG.warning("failed to fetch %s", url)
        return
    with open(f"{now:%Y%m%d_%H}00.gz", "wb") as fh:
        fh.write(resp.content)
    subprocess.call(["gunzip", f"{now:%Y%m%d_%H}00.gz"])
    subprocess.call(["mv", f"{now:%Y%m%d_%H}00", testfn])
    LOG.warning("backfilled %s", testfn)


if __name__ == "__main__":
    main()



"""Test mcd workflow."""

import pytest

from pywwa.testing import get_example_file
from pywwa.workflows import mcd


@pytest.mark.parametrize("database", ["postgis"])
def test_simple(cursor):
    """Test MCD processing."""
    mcd.real_process(cursor, get_example_file("MCD.txt"))



"""For testing."""

# 3rd party
import pytest

from pywwa import CTX, CTX_DEFAULTS
from pywwa.database import get_dbconnc

# Dragons
# pytest + twisted + click == too much magic for poor me
# So, this effectively disables reactor.stop() from working
CTX_DEFAULTS["shutdown_delay"] = 1800


@pytest.fixture(autouse=True)
def setup_and_teardown():
    """Something to fix the hacky global namespace mucking I do."""
    CTX.update(CTX_DEFAULTS)
    yield
    CTX.update(CTX_DEFAULTS)


@pytest.fixture(scope="function")
def cursor(database):
    """Return a disposable database cursor."""
    pgconn, cursor = get_dbconnc(database)
    yield cursor
    pgconn.close()
